from __future__ import annotations

import functools
import os
from abc import ABC, abstractmethod
from typing import (Any, Deque, Dict, FrozenSet, Generic, Iterable, List,
                    Optional, Sequence, Tuple, TypeVar)

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from ..activelearning.base import ActiveLearner
from ..activelearning.estimator import Estimator
from ..utils.func import (all_subsets, intersection, list_unzip3,
                          not_in_supersets, powerset, union)
from .base import AbstractEstimator, Estimate
from .rcapture import AbundanceEstimator, _check_R

try:
    import rpy2.robjects as ro  # type: ignore
    from rpy2.robjects import pandas2ri  # type: ignore
    from rpy2.robjects.conversion import localconverter  # type: ignore
    from rpy2.robjects.packages import importr  # type: ignore
except ImportError:
    R_AVAILABLE = False
else:
    R_AVAILABLE = True

DT = TypeVar("DT")
VT = TypeVar("VT")
KT = TypeVar("KT")
LT = TypeVar("LT")
RT = TypeVar("RT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")
_T = TypeVar("_T")
_U = TypeVar("_U")


class RaschEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "RaschEstimator"
    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "rasch_estimate_bootstrap.R")
        R["source"](r_script_file)

    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        interaction_cols = {
            f"h{i-1}": len(all_subsets(learner_set, i, i)) 
            for i in range(2, len(all_learners))
        }
        final_row = {
            **learner_cols,
            **interaction_cols,
            **count_col
        }
        return final_row

    def get_occasion_history(self, 
                             estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                             label: LT) -> pd.DataFrame:
        contingency_sets = self.get_contingency_sets(estimator, label)
        learner_keys = union(*contingency_sets.keys())
        rows = {i:
            self.rasch_row(pair, learner_keys, True)
            for (i, pair) in enumerate(contingency_sets.items())
        }
        df = pd.DataFrame.from_dict(# type: ignore
            rows, orient="index")
        self.matrix_history.append(df)
        return df

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df = self.get_occasion_history(estimator, label)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(df)
            abundance_r = ro.globalenv["rasch.single"]
            r_df = abundance_r(df_r)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    
    def calculate_abundance(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        estimate_missing = res_df.values[0,0]
        estimate_error = res_df.values[0,1]
        total_found = estimator.env.labels.document_count(label)
        return (total_found + estimate_missing), estimate_error * 2

    def all_estimations(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT) -> Sequence[Tuple[str, float, float]]:
        return []

    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        abundance, error = self.calculate_abundance(learner, label)
        return abundance, abundance-error, abundance + error

class NonParametricRasch(RaschEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "RaschNonParamatric"
    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "rasch_estimate_bootstrap.R")
        R["source"](r_script_file)

    

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df = self.get_occasion_history(estimator, label)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(df)
            abundance_r = ro.globalenv["rasch.nonparametric"]
            r_df = abundance_r(df_r)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_estimate(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        horizon = res_df.values[0,0]
        horizon_lowerbound = res_df.values[0,1]
        horizon_upperbound = res_df.values[0,2]
        return horizon, horizon_lowerbound, horizon_upperbound
    
    def __call__(self, 
        learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        estimate, lower_bound, upper_bound = self.calculate_estimate(learner, label)
        return estimate, lower_bound, upper_bound

class ParametricRasch(NonParametricRasch[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "RaschParametric"
    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df = self.get_occasion_history(estimator, label)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(df)
            abundance_r = ro.globalenv["rasch.parametric"]
            r_df = abundance_r(df_r)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

class SingleInteractionRaschParametric(ParametricRasch[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "ParametericRaschInteractionsCombined"
    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        interaction_col = {"h1": len(
            all_subsets(
                learner_set, 
                2, len(all_learners) - 1)) 
        }
        final_row = {
            **learner_cols,
            **interaction_col,
            **count_col
        }
        return final_row

class SingleInteractionRaschParametric2(
        ParametricRasch[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "ParametricRaschOnly2ndOrder"
    
    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
                  all_learners: FrozenSet[int],
                  positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        interaction_col = {"h1": len(
            all_subsets(
                learner_set, 
                2, 2)) 
        }
        final_row = {
            **learner_cols,
            **interaction_col,
            **count_col
        }
        return final_row

class EMRasch(NonParametricRasch[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "EMRasch"
    def __init__(self, neg_label: LT) -> None:
        super().__init__()
        self.neg_label = neg_label

    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "rasch_em.R")
        R["source"](r_script_file)

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df_pos = self.get_occasion_history(estimator, label)
        df_neg = self.get_occasion_history(estimator, self.neg_label)
        dataset_size = len(estimator.env.dataset)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_pos_r = ro.conversion.py2rpy(df_pos)
            df_neg_r = ro.conversion.py2rpy(df_neg)
            abundance_r = ro.globalenv["rasch.em"]
            r_df = abundance_r(df_pos_r, df_neg_r, dataset_size)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_estimate(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        horizon = res_df.values[0,0]
        return horizon, horizon, horizon
    
    def __call__(self, 
        learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        estimate, lower_bound, upper_bound = self.calculate_estimate(learner, label)
        return estimate, lower_bound, upper_bound

class EMRaschCombined(NonParametricRasch[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "EMRaschCombined"
    def get_contingency_sets_negative(self, 
                                      estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                                      label: LT) -> Dict[FrozenSet[int], FrozenSet[KT]]:
        learner_sets = {
                learner_key: (
                    frozenset(learner.env.labeled).
                        difference(
                            learner.env.labels.get_instances_by_label(label))
                )
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: intersection(*[learner_sets[key] for key in combination])
            for combination in key_combinations
            if len(combination) >= 1
        }
        filtered_result = not_in_supersets(result)
        return filtered_result

    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        positive_col = {"positive": int(positive)}
        interaction_cols = {
            f"h{i-1}": len(all_subsets(learner_set, i, i)) 
            for i in range(2, len(all_learners))
        }
        pos_learner_cols = {
            f"learner_{learner_key}-positive": (
                int(learner_key in learner_set) if positive else 0)
            for learner_key in all_learners
        }
        interaction_pos_cols = {
            f"h{i-1}-positive": (
                len(all_subsets(learner_set, i, i)) if positive else 0)
            for i in range(2, len(all_learners))
        }
        final_row = {
            **learner_cols,
            **positive_col,
            **pos_learner_cols,
            **interaction_cols,
            **interaction_pos_cols,
            **count_col
        }
        return final_row

    def get_occasion_history(self, 
                             estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                             label: LT) -> pd.DataFrame:
        contingency_sets_pos = self.get_contingency_sets(estimator, label)
        contingency_sets_neg = self.get_contingency_sets_negative(estimator, label)
        learner_keys = union(*contingency_sets_pos.keys())
        
        rasch_pos_func = functools.partial(
            self.rasch_row, all_learners=learner_keys, positive = True)
        rasch_neg_func = functools.partial(
            self.rasch_row, all_learners=learner_keys, positive = False)
        
        
        rows_pos = list(map(rasch_pos_func, contingency_sets_pos.items()))
        rows_neg = list(map(rasch_neg_func, contingency_sets_neg.items()))

        rows = dict(enumerate(rows_pos + rows_neg))
        
        df = pd.DataFrame.from_dict(# type: ignore
            rows, orient="index")
        self.matrix_history.append(df)
        return df

    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "rasch_em_comb.R")
        R["source"](r_script_file)

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df_pos = self.get_occasion_history(estimator, label)
        dataset_size = len(estimator.env.dataset)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_pos_r = ro.conversion.py2rpy(df_pos)
            abundance_r = ro.globalenv["rasch.em.horizon"]
            r_df = abundance_r(df_pos_r, dataset_size)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_estimate(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        horizon = res_df.values[0,0]
        return horizon, horizon, horizon
    
    def __call__(self, 
        learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        estimate, lower_bound, upper_bound = self.calculate_estimate(learner, label)
        return Estimate(estimate, lower_bound, upper_bound)

class EMAlternative(EMRaschCombined[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        positive_col = {"positive": int(positive)}
        interaction_cols = {
            f"h{i-1}": len(all_subsets(learner_set, i, i)) 
            for i in range(2, len(all_learners))
        }
        interaction_pos_cols = {
            f"h{i-1}-positive": (
                len(all_subsets(learner_set, i, i)) if positive else 0)
            for i in range(2, len(all_learners))
        }
        final_row = {
            **learner_cols,
            **positive_col,
            **interaction_cols,
            **interaction_pos_cols,
            **count_col
        }
        return final_row

class EMRaschRidge(EMRaschCombined[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df_pos = self.get_occasion_history(estimator, label)
        dataset_size = len(estimator.env.dataset)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_pos_r = ro.conversion.py2rpy(df_pos)
            abundance_r = ro.globalenv["rasch.ridge.em.horizon"]
            r_df = abundance_r(df_pos_r, dataset_size)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

class EMRaschCombinedBootstrap(EMRaschCombined[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df_pos = self.get_occasion_history(estimator, label)
        dataset_size = len(estimator.env.dataset)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_pos_r = ro.conversion.py2rpy(df_pos)
            abundance_r = ro.globalenv["rasch.em.bootstrap.horizon"]
            r_df = abundance_r(df_pos_r, dataset_size)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_estimate(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        horizon = res_df.values[0,0]
        lowerbound = res_df.values[0,1]
        upperbound = res_df.values[0,2]
        return horizon, lowerbound, upperbound

class EMRaschRidgeCombinedBootstrap(EMRaschCombined[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df_pos = self.get_occasion_history(estimator, label)
        dataset_size = len(estimator.env.dataset)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_pos_r = ro.conversion.py2rpy(df_pos)
            abundance_r = ro.globalenv["rasch.ridge.em.horizon.parametric"]
            r_df = abundance_r(df_pos_r, dataset_size)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_estimate(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        horizon = res_df.values[0,0]
        lowerbound = res_df.values[0,1]
        upperbound = res_df.values[0,2]
        return horizon, lowerbound, upperbound



import functools
import itertools

from typing import Any, Dict, FrozenSet, Iterable, List, Mapping, TypeVar, Callable,  Tuple, Optional, Sequence

_T = TypeVar("_T")
_U = TypeVar("_U")
_V = TypeVar("_V")

def hn(val: Optional[_T], default: _T) -> _T:
    if val is not None:
        return val
    return default

def prepend_keys(mapping: Mapping[str, _T], prefix: str) -> Mapping[str, _T]:
    return {f"{prefix}{key}": value for key, value in mapping.items()}

def powerset(iterable: Iterable[_T]) -> FrozenSet[FrozenSet[_T]]:
    """Calculates the powerset of an interable.


    Parameters
    ----------
    iterable : Iterable[_T]
        An iterable of which we want to calculate the powerset

    Returns
    -------
    FrozenSet[FrozenSet[_T]]
        Returns a frozenset of frozensets containing all elements
        of the powerset

    Examples
    --------
    
    Usage
    >>> powerset([1,2,3])
    >>> # frozenset({() (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)})
    >>> # all elements are also frozensets
    """
    s = list(iterable)
    result = itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s)+1))
    return frozenset(map(frozenset, result))  # type: ignore



def not_in_supersets(
        contingency: Dict[FrozenSet[_T], FrozenSet[_U]]
        ) -> Dict[FrozenSet[_T], FrozenSet[_U]]:
    """This function filters out all values that also exist in supersets
    in the dictionary for the values that exist in
    all pairs in the mapping key sets to value sets

    TODO: Make a clear description for this function

    Parameters
    ----------
    contingency : Dict[FrozenSet[_T], FrozenSet[_U]]
        A dictionary that maps key set to sets of values

    Returns
    -------
    Dict[FrozenSet[_T], FrozenSet[_U]]
        A dictionary that maps sets to sets of of values.
        The values of that belong to a key set only contain
        values that do not exist in a superset of the keyset.
    """    
    ret_dict: Dict[FrozenSet[_T], FrozenSet[_U]] = {}
    sets = frozenset(contingency.keys())
    for key_set in sets:
        strict_supersets = frozenset(filter(
            lambda s: s.issuperset(key_set) and s != key_set,
            sets))
        in_supersets: FrozenSet[_U] = frozenset()
        if len(strict_supersets) > 0:
            in_supersets = union(*map(lambda k: contingency[k], strict_supersets))
        ret_dict[key_set] = contingency[key_set].difference(in_supersets)
    return ret_dict


def flatten_dicts(*dicts: Mapping[_T, _U]) -> Dict[_T, _U]:
    """Recursive function that combines a list of dictionaries
    into a single dictionary. When a key exists multiple times, 
    the last one is preserved.

    Parameters
    ----------
    dicts
        A sequence dicts that should be combined

    Returns
    -------
    Dict[_T, _U]
        A single dictionary combining all given dictionaries.
    """    
    try:
        head, *tail = dicts
        return {**head, **flatten_dicts(*tail)}
    except ValueError:
        return {}

def intersection(*sets: FrozenSet[_T]) -> FrozenSet[_T]:
    """Return the intersection of all sets in the parameters
    
    Parameters
    ----------
    sets
        A sequence of frozensets that should be combined

    Returns
    -------
    FrozenSet[_T]
        A frozenset that contains the intersection of all the sets
    """    
    try:
        head, *tail = sets
        return head.intersection(*tail)
    except ValueError:
        return frozenset()

def union(*sets: FrozenSet[_T]) -> FrozenSet[_T]:
    """Return the union of all sets in the parameters
    
    Parameters
    ----------
    sets
        A sequence of frozensets that should be combined

    Returns
    -------
    FrozenSet[_T]
        A frozenset that contains the union of all the sets
    """    
    try:
        head, *tail = sets
        return head.union(*tail)
    except ValueError:
        return frozenset()
def find_subsets(s: FrozenSet[_T], n: int) -> FrozenSet[FrozenSet[_T]]:
    
    subsets_n = frozenset(
        map(frozenset, itertools.combinations(s, n))) # type: ignore
    return subsets_n # type: ignore

def all_subsets(fset: FrozenSet[_T], 
                n_min: int, n_max: int) -> FrozenSet[FrozenSet[_T]]:
    """Find all subsets of `fset` with size `n_min` to and including 
    `n_max`

    Parameters
    ----------
    fset : FrozenSet[_T]
        The input set
    n_min : int
        The mininum size of the subsets
    n_max : int
        The maximum size of the subsets

    Returns
    -------
    FrozenSet[FrozenSet[_T]]
        A set of subsets
    """    
    subsets = (find_subsets(fset, n) for n in range(n_min, n_max + 1))
    result = union(*subsets)
    return result

def expandgrid(**itrs: Iterable[_T]) -> Dict[str, List[_T]]:
    """Port of the expand.grid function from R

    Parameters
    ----------
    itrs: Keyword variables with an iterable as argument

    Returns
    -------
    Dict[str, List[_T]]
        A dictionary with as values all possible combinations in a gridlike 
        ordering
    """
    variables = list(itrs.keys())
    iterators = list(itrs.values())
    product = list(itertools.product(*iterators))
    return {variable: [x[i] for x in product] for i, variable in enumerate(variables)}

def mapsnd(func: Callable[[_U], _V]) -> Callable[[_T, _U], Tuple[_T, _V]]:
    """This function converts a function with signature `u -> v` to a function
    with signature `(t, u) -> (t, v)`. 
    
    That is, a function that is applied to the second
    element of a tuple and where the first element is preserved.

    Parameters
    ----------
    func 
        The original function

    Returns
    -------
    Callable[[_T, _U], Tuple[_T, _V]]
        The new function

    Examples
    --------
    Usage

    >>> def plustwo(number: int) -> int:
    ...     return number + 2
    >>> tuple_variable = (2, 3)
    >>> mapped_p2 = mapsnd(plustwo)
    >>> result = mapped_p2(tuple_variable)
    >>> print(result)
    >>> # (2, 5)
    """    
    @functools.wraps(func)
    def wrap(fst: _T, snd: _U) -> Tuple[_T, _V]:
        result = func(snd)
        return (fst, result)
    return wrap

def all_equal(iterable: Iterable[Any]) -> bool:
    """Checks if all the elements of an iterable are the same

    Parameters
    ----------
    iterable
        An iterable of generic type

    Returns
    -------
    bool
        True if all elements are the same
    """    
    g = itertools.groupby(iterable)
    return next(g, True) and not next(g, False) # type: ignore

def list_unzip(iterable: Iterable[Tuple[_T, _U]]) -> Tuple[Sequence[_T], Sequence[_U]]:
    """Unzips an iterable of tuples of two elements, and returns a tuple of two lists,
    where the first contains all the first elements of the tuples and the latter the second
    elements

    Parameters
    ----------
    iterable
        An iterable of tuples of indiscriminate type

    Returns
    -------
    Tuple[Sequence[_T], Sequence[_U]]
        A tuple of two sequences: 
        
        - The first contains all first elements (of type `_T`)
        - The second contains all second elements (of type `_U`)

    Examples
    --------
    Usage

    >>> tuple_list = [(1, "a"), (2, "b"), (3, "c")]
    >>> idx, abcs = list_unzip(tuple_list)
    >>> print(idx)
    >>> # [1, 2, 3]
    >>> print(abcs)
    >>> # ["a", "b", "c"]
    """    
    try:
        fst_iter, snd_iter = zip(*iterable)
        fst_sequence, snd_sequence = list(fst_iter), list(snd_iter)
    except ValueError:
        return [], [] # type: ignore
    return fst_sequence, snd_sequence # type: ignore

def list_unzip3(
        iterable: Iterable[Tuple[_T, _U, _V]]) -> Tuple[Sequence[_T], Sequence[_U], Sequence[_V]]:
    """Unzips an iterable of tuples of three elements, and returns a tuple of three lists,
    where the first contains all the first elements of the tuples and the latter the second
    elements, etc.

    Parameters
    ----------
    iterable
        An iterable of tuples of indiscriminate type

    Returns
    -------
    Tuple[Sequence[_T], Sequence[_U], Sequence[_V]]
        A tuple of three sequences: 
        
        - The first contains all first elements (of type `_T`)
        - The second contains all second elements (of type `_U`)
        - The third contains all third elements (of type `_V`)

    Examples
    --------
    Usage

    >>> tuple_list = [(1, "a", 20), (2, "b", 40), (3, "c", 60)]
    >>> idx, abcs, ints = list_unzip3(tuple_list)
    >>> print(idx)
    >>> # [1, 2, 3]
    >>> print(abcs)
    >>> # ["a", "b", "c"]
    >>> print(ints)
    >>> # [20, 40, 60]
    """    
    try:
        fst_sequence, snd_sequence, trd_sequence = zip(*iterable)  # type ignore
    except ValueError:
        return [], [], [] #type: ignore
    return list(fst_sequence), list(snd_sequence), list(trd_sequence) # type: ignore

def filter_snd_none_zipped(iter: Iterable[Tuple[_T, Optional[_U]]]) -> Tuple[Sequence[_T], Sequence[_U]]:
    """Filter a list of tuples where the second element is not None

    Parameters
    ----------
    iter : Iterable[Tuple[_T, Optional[_U]]]
        A list of tuples where the second element may be `None`

    Returns
    -------
    Tuple[Sequence[_T], Sequence[_U]]
        A tuple of two lists of equal length based on all tuples where the 
        second element was not `None`:

        - A list of elements with type `_T`
        - A list of elements with type `_U`

    See Also
    --------
    filter_snd_none
        This function works similarly, but here the 
        iterable is not in `zipped` form.
    """    
    filtered = filter(lambda x: x[1] is not None, iter)
    return list_unzip(filtered) # type: ignore

def filter_snd_none(fst_iter: Iterable[_T], 
                snd_iter: Iterable[Optional[_U]]
                ) -> Tuple[Sequence[_T], Sequence[_U]]:
    """Filter two iterables of equal length for pairs where the second element is not None
    
    Parameters
    ----------
    fst_iter
        The first iterable
    snd_iter
        The second iterable where elements may be `None`.

    Returns
    -------
    Tuple[Sequence[_T], Sequence[_U]]
         A tuple of two lists of equal length based on all pairs based on index where the 
        element in the second list was not `None`:

        - A list containing all eligble elements from `fst_iter`
        - A list containing all eligble elements from `snd_iter`

    Examples
    --------

    Usage:

    >>> a = [1,2,3,4,5,6,7,8,9,10]
    >>> b = ["a", None, "c", "d", "e", "f", "g", None,"i", "j"]
    >>> a_filtered, b_filtered = filter_snd_none(a,b)
    >>> print(a_filtered)
    >>> # [1, 3, 4, 5, 6, 7, 9, 10]
    >>> print(b_filtered)
    >>> # ["a", "c", "d", "e", "f", "g", "i", "j"]

    See Also
    --------
    filter_snd_none_zipped : 
        This function works similarly, but here the 
        iterable is in `zipped` form, i.e., a list of tuples. 
    """    
    zipped: Iterable[Tuple[_T, _U]] = filter(lambda x: x[1] is not None, zip(fst_iter, snd_iter)) # type: ignore
    return list_unzip(zipped)

def sort_on(index: int, seq: Sequence[Tuple[_T, _U]]) -> Sequence[Tuple[_T, _U]]:
    """A function that allows you to sort a sequence on the `n-th` element of a tuple
    it contains. This function orders in descending order (high to low).

    Parameters
    ----------
    index : int
        The index of the tuple element on which you desire to sort
    seq 
        The sequence of tuples

    Returns
    -------
    Sequence[Tuple[_T, _U]]
        A sorted sequence of tuples
    """    
    return sorted(seq, key=lambda k: k[index], reverse=True) # type: ignore

def multisort(seq: Sequence[Tuple[_T, Sequence[_U]]]) -> Sequence[Sequence[Tuple[_T,_U]]]:
    len_scores = len(seq[0][1])
    unpacked = [[(key, scores[i]) for (key, scores) in seq] for i in range(len_scores)]
    sorted_unpacked = [sort_on(1, sublist) for sublist in unpacked]
    return sorted_unpacked

def fst(tup: Tuple[_T, Any]) -> _T:
    first_element = tup[0]
    return first_element



import collections
import itertools
import multiprocessing as mp
import warnings
from dataclasses import dataclass
from logging import currentframe
from os import spawnlp
from typing import Any, Deque, Generic, List, Optional, Sequence, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
from numba import (int32, jit, prange, set_num_threads, typed,  # type: ignore
                   types)
from numba.experimental import jitclass
from numba.typed import List as NumbaList
from numpy.linalg import LinAlgError

from ..activelearning.base import ActiveLearner
from ..activelearning.estimator import Estimator
from ..analysis.statistics import EstimationModelStatistics
from ..estimation.base import Estimate
from ..typehints.typevars import DT, KT, LT, RT, VT
from .rasch import EMRaschCombined

spec = [
    ("positive", int32[:]),
    ("estimate", int32[:]),
    ("negative", int32[:]),
    ("observed", int32[:]),
    ("positive_estimate", int32[:]),
    ("negative_estimate", int32[:]),
    ("positive_observed", int32[:]),
    ("negative_observed", int32[:]),
]

@jitclass(spec)
class Masks:
    def __init__(self, 
                 positive: npt.NDArray[Any], 
                 estimate: npt.NDArray[Any], 
                 negative: npt.NDArray[Any], 
                 observed: npt.NDArray[Any], 
                 positive_estimate: npt.NDArray[Any], 
                 negative_estimate: npt.NDArray[Any], 
                 positive_observed: npt.NDArray[Any], 
                 negative_observed: npt.NDArray[Any]):
        self.positive = positive
        self.estimate = estimate
        self.negative = negative
        self.observed = observed
        self.positive_estimate = positive_estimate
        self.negative_estimate = negative_estimate
        self.positive_observed = positive_observed
        self.negative_observed = negative_observed

def create_mask(poses: List[bool], estes: List[bool]) -> Masks:
    def convert(bool_list: Sequence[bool]) -> npt.NDArray[Any]:
        converted = np.array([i for (i,b) in enumerate(bool_list) if b], dtype=np.int32)
        return converted
    positive = convert(poses)
    estimate = convert(estes)
    neges = [not(m) for m in poses]
    obses = [not(m) for m in estes]
    negative = convert(neges)
    observed = convert(obses)
    positive_estimate = convert(NumbaList([est & pos for est, pos in zip(estes, poses)]))
    negative_estimate = convert(NumbaList([est & neg for est, neg in zip(estes, neges)]))
    positive_observed = convert(NumbaList([obs & pos for obs, pos in zip(obses, poses)]))
    negative_observed = convert(NumbaList([obs & neg for obs, neg in zip(obses, neges)]))
    return Masks(positive, estimate, 
                 negative, observed, 
                 positive_estimate, negative_estimate, 
                 positive_observed, negative_observed)

@jit(nopython=True, cache=True) # type: ignore
def l2(b0: npt.NDArray[Any],
       design_mat: npt.NDArray[Any],
       counts: npt.NDArray[Any],
       lam: float = 0.0,
       tolerance: float = 1e-8,
       max_it: int = 10000,
       epsilon: float = 0.1
       ) -> npt.NDArray[Any]:
    counts = counts + epsilon
    current_tolerance = 1
    b = b0
    it = 0
    while current_tolerance > tolerance and it < max_it:
        it = it + 1
        # mu <- exp(D %*% b0)
        mu = np.exp(design_mat @ b0)

        # db <- crossprod(D, Freq) - crossprod(D, mu) - 2 * c(0, lambda * b0[-1])
        all_but_first_b0: npt.NDArray[Any] = b0[1:]
        lbpart: npt.NDArray[Any] = 2 * np.concatenate(
            (
                np.array([0.0]),
                (lam * all_but_first_b0)
        ))
        design_mat_t = design_mat.conj().transpose()
        db = design_mat_t @ counts - design_mat_t @ mu - lbpart

        # I <- crossprod(D, diag(c(mu))) %*% D + 2 * diag(c(0, rep(lambda, length(b0) - 1)))
        repeat_count = b0.shape[0] - 1
        other_lb_part = np.hstack((np.array([0.0]), np.repeat(lam, repeat_count)))
        I = (design_mat_t @ np.diag(mu)) @ design_mat + \
            2 * np.diag(other_lb_part)
        try:
            invI = np.linalg.inv(I)
        except Exception: # LinAlgException
            # Change lambda if the matrix is not invertible
            lam = 1e-6 if lam == 0 else 10 * lam
        else:
            # If inversion succeeded, calculate new coefficients b
            b = b0 + invI @ db
            current_tolerance = np.sum(np.abs(b0 - b))
            b0 = b
    return b


@jit(nopython=True, cache=True)  # type: ignore
def calc_deviance(y: npt.NDArray[Any], y_hat: npt.NDArray[Any]) -> float:
    pos_y_idx = np.where(y > 0)
    pos_y = y[pos_y_idx]
    deviance = (
        2 * np.sum(pos_y * np.log(pos_y / y_hat[pos_y_idx])) -
        2 * np.sum(y - y_hat))
    return deviance


def add_missing_count_rows(freq_df: pd.DataFrame, n_pos: float, n_neg: float) -> pd.DataFrame:
    pos_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_pos, "positive": 1}
    }
    neg_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_neg, "positive": 0}
    }
    new_freq_df = (
        freq_df
        .append(pos_dict, ignore_index=True)
        .append(neg_dict, ignore_index=True)
    )
    return new_freq_df


def l2_format(freq_df: pd.DataFrame, learner_cols: Sequence[str]) -> pd.DataFrame:
    order_cols = ["positive", *learner_cols]
    df = freq_df.sort_values(order_cols)
    df.insert(0, "intercept", 1)
    return df

@jit(nopython=True, cache=True)  # type: ignore
def parameter_ci(estimate: float, se: float) -> Tuple[float, float]:
    var_min = estimate - 2 * se
    var_max = estimate + 2 * se
    return var_max, var_min

@jit(nopython=True, cache=True)  # type: ignore
def calculate_estimate(intercept: float, pos: float) -> float:
    return np.exp(intercept + pos)

def calculate_ci_rasch(intercept: float,
                       positive: float,
                       intercept_se: float,
                       positive_se: float) -> Tuple[float, float]:
    intercepts = parameter_ci(intercept, intercept_se)
    positive_par = parameter_ci(positive, positive_se)
    combinations = itertools.product(intercepts, positive_par)
    results = list(itertools.starmap(calculate_estimate, combinations))
    min_result = min(results)
    max_result = max(results)
    return min_result, max_result

@jit(nopython=True, cache=True)  # type: ignore
def initial_fit(n_not_read: float, design_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
    not_read = np.array([n_not_read])
    b0 = np.hstack(
        (
            np.log(not_read),
            np.repeat(0.0, design_mat.shape[1] - 1)
    ))
    return b0

@jit(nopython=True, cache=True)  # type: ignore
def rasch_glm(design_mat: npt.NDArray[Any],
              efit: npt.NDArray[Any],
              counts: npt.NDArray[Any],
              b0: npt.NDArray[Any],
              masks: Masks,
              epsilon: float = 0.0
              ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    obs_counts = counts[masks.observed]
    beta = l2(b0, design_mat, efit, epsilon=epsilon)
    mfit = np.exp(design_mat @ beta)
    obs_fitted = mfit[masks.observed]
    deviance = calc_deviance(obs_counts, obs_fitted)
    return beta, mfit, deviance

@jit(nopython=True, cache=True) # type: ignore
def glm(design_mat: npt.NDArray[Any], counts: npt.NDArray[Any], epsilon=0.1) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    b0 = np.repeat(1.0, design_mat.shape[1])
    beta = l2(b0, design_mat, counts, epsilon=epsilon)
    mfit = np.exp(design_mat @ beta)
    deviance = calc_deviance(counts, mfit)
    return beta, mfit, deviance

def remove_pos_cells(df: pd.DataFrame) -> pd.DataFrame:
    pos_rows = df[df["positive"] > 0] 
    not_pos_cols = pos_rows.filter(regex="^((?!positive).)*$")
    return not_pos_cols

@jit(nopython=True, parallel=True, cache=True) # type: ignore
def pos_rasch_numpy(design_mat: npt.NDArray[Any], 
                    obs_counts: npt.NDArray[Any], 
                    total_found: int, max_it: int = 2000
                    ) -> Tuple[float, float, float, npt.NDArray[Any], npt.NDArray[Any], float, npt.NDArray[Any]]:
    beta, mfit, deviance = glm(design_mat, obs_counts, epsilon = 0.1)
    estimate = np.exp(beta[0])
    horizon_estimate = int(np.round(total_found + estimate))
    fitted: npt.NDArray[Any] = np.concatenate((np.array([estimate]), mfit))
    p_vals = fitted / np.sum(fitted)
    
    if not np.isnan(estimate):
        deviances = np.zeros(max_it)
        betas = np.zeros((max_it, design_mat.shape[1]))
        mfits = np.zeros((max_it, design_mat.shape[0]))
        multinomial_fits: npt.NDArray[Any] = np.random.multinomial(horizon_estimate, p_vals, max_it)
            
        for idx in prange(max_it):
            multinom = multinomial_fits[idx,1:].copy()
            i_beta, i_mfit, i_deviance = glm(design_mat, multinom, epsilon=0.1)
            betas[idx,:] = i_beta
            mfits[idx,:] = i_mfit
            deviances[idx] = i_deviance
        
        estimates = total_found + np.exp(betas[:,0])
        low = np.percentile(estimates, 2.5)
        up = np.percentile(estimates, 97.5)
        point = np.percentile(estimates, 50)    
        return point, low, up, beta, mfit, deviance, estimates
    return estimate, estimate, estimate, beta, mfit, deviance, np.array([estimate])




def rasch_estimate_only_pos(freq_df: pd.DataFrame, multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    learner_cols = list(freq_df.filter(regex=r"^learner"))
    df_formatted = remove_pos_cells(l2_format(freq_df, learner_cols))   
    design_mat = (
        df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
            .values) # type: ignore
    design_mat = design_mat.astype("float64")
    obs_counts: npt.NDArray[Any] = df_formatted["count"].values # type: ignore
    total_found: int = int(np.sum(obs_counts))
    
    point, low, up, beta, mfit, deviance, estimates = pos_rasch_numpy(
        design_mat, obs_counts, total_found, max_it=multinomial_size)
    estimate = Estimate(point, low, up)
    stats = EstimationModelStatistics(beta, mfit, deviance, estimates)
    return estimate, stats





@jit(nopython=True, cache=True)  # type: ignore
def rasch_em(design_mat: npt.NDArray[Any],
             b0: npt.NDArray[Any],
             counts: npt.NDArray[Any],
             n_dataset: float,
             masks: Masks,
             proportion: float,
             tolerance: float = 1e-5,
             max_it=1000) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    em_b0 = b0.copy()
    em_counts = counts.copy()
    orig_counts = em_counts[masks.observed]
    n_not_read = n_dataset - np.sum(orig_counts)
    
    n_pos = proportion * n_not_read
    n_neg = (1 - proportion) * n_not_read
    efit = em_counts.copy()

    efit[masks.positive_estimate] = n_pos
    efit[masks.negative_estimate] = n_neg

    beta, mfit, deviance = rasch_glm(design_mat, efit, counts, em_b0, masks)
    current_tolerance = 1
    # Expectation Maximization
    it = 0
    while current_tolerance > tolerance:
        old_deviance = deviance
        efit[masks.estimate] = mfit[masks.estimate] * n_not_read / np.sum(mfit[masks.estimate])
        beta, mfit, deviance = rasch_glm(design_mat, efit, counts, beta, masks)
        current_tolerance = old_deviance - deviance
        if it > max_it or np.any(np.isnan(mfit)):
            #warnings.warn(
            #    f"Exceeded maximum EM iterations ({max_it}). Estimate might not be accurate")
            return beta, mfit, deviance
    return beta, mfit, deviance

@jit(nopython=True, cache=True)  # type: ignore
def rasch_bf(design_mat: npt.NDArray[Any],
             b0: npt.NDArray[Any],
             counts: npt.NDArray[Any],
             n_dataset: float,
             masks: Masks) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], npt.NDArray[Any]]:
    em_b0 = b0.copy()
    em_counts = counts.copy()
    orig_counts = em_counts[masks.observed]
    n_not_read = n_dataset - np.sum(orig_counts)
    
    ar_size = int(n_not_read) + 1

    betas = np.zeros((ar_size, design_mat.shape[1]))
    mfits = np.zeros((ar_size, design_mat.shape[0]))
    deviances = np.zeros((ar_size))
            
    efit = em_counts.copy()

    for n_pos in prange(0, n_not_read + 1):
        n_neg = n_not_read - n_pos
        efit[masks.positive_estimate] = n_pos
        efit[masks.negative_estimate] = n_neg
        
        i_beta, i_mfit, i_deviance = rasch_glm(design_mat, efit, counts, em_b0, masks)
        betas[n_pos,:] = i_beta
        mfits[n_pos,:] = i_mfit
        deviances[n_pos] = i_deviance
    return betas, mfits, deviances

@jit(nopython=True, cache=True)
def select_best_model(betas: npt.NDArray[Any], mfits: npt.NDArray[Any], deviances: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float]:
    best_beta = np.zeros((betas.shape[1],))
    best_mfit = np.zeros((mfits.shape[1],))
    best_deviance = 2000000.0
    for idx in range(betas.shape[0]):
        current_beta = betas[idx,:]
        current_mfit = mfits[idx,:]
        current_deviance = deviances[idx]
        if current_deviance < best_deviance:
            best_deviance = current_deviance
            best_beta = current_beta
            best_mfit = current_mfit
    return best_beta, best_mfit, best_deviance

@jit(nopython=True, parallel=True, cache=True) # type: ignore
def rasch_numpy(design_mat: npt.NDArray[Any],
                counts: npt.NDArray[Any],
                n_dataset: float,
                masks: Masks,
                proportions: npt.NDArray[Any] = np.array([0.001, 0.01, 0.1, 0.5, 0.75, 0.9]),
                tolerance: float = 1e-5,
                max_it=1000
                ) -> Tuple[npt.NDArray[Any], npt.NDArray[Any], float, float]:
    deviances = np.zeros(proportions.shape)
    betas = np.zeros((proportions.shape[0], design_mat.shape[1]))
    mfits = np.zeros((proportions.shape[0], design_mat.shape[0]))
    num_proportions = proportions.shape[0]
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    for idx in prange(num_proportions):
        proportion = proportions[idx]
        beta, mfit, deviance = rasch_em(design_mat, b0,  
                                        counts, n_dataset, masks, 
                                        proportion, tolerance, max_it)
        betas[idx,:] = beta
        mfits[idx,:] = mfit
        deviances[idx] = deviance
    best_beta = np.zeros((design_mat.shape[1],))
    best_mfit = np.zeros((design_mat.shape[0],))
    best_deviance = 2000000.0
    best_proportion = 1.0
    for idx in range(num_proportions):
        current_beta = betas[idx,:]
        current_mfit = mfits[idx,:]
        current_deviance = deviances[idx]
        if current_deviance < best_deviance:
            best_deviance = current_deviance
            best_beta = current_beta
            best_mfit = current_mfit
            best_proportion = proportions[idx]
    return best_beta, best_mfit, best_deviance, best_proportion

@jit(nopython=True, parallel=True, nogil=True, cache=True) # type: ignore
def rasch_parallel(design_mat: npt.NDArray[Any],
                   b0: npt.NDArray[Any], 
                   rounded_mfits: npt.NDArray[Any], 
                   n_dataset: int,
                   proportion: float,
                   masks: Masks):
    # Placeholders for results
    deviances = np.zeros(rounded_mfits.shape[0])
    betas = np.zeros((rounded_mfits.shape[0], design_mat.shape[1]))
    mfits = np.zeros((rounded_mfits.shape[0], design_mat.shape[0]))
    num_mfits = rounded_mfits.shape[0]

    # Parallel execution of Parametric Bootstrap
    for idx in prange(num_mfits):
        mfit_sample = rounded_mfits[idx,:].copy()
        b0_copy = b0.copy()
        beta, mfit, deviance = rasch_em(
            design_mat, b0_copy, mfit_sample,
            n_dataset, masks, proportion) 
        betas[idx,:] = beta
        mfits[idx,:] = mfit
        deviances[idx] = deviance
    
    # Reordering the results
    results = NumbaList()
    for idx in range(num_mfits):
        result = (betas[idx,:], mfits[idx,:], deviances[idx])
        results.append(result)
    return results
        
def rasch_estimate_parametric_approx(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 2000) -> Tuple[float, float, float]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore

    only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    proportion: float = only_pos_estimate / (n_dataset - np.sum(counts[masks.observed]))
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = rasch_em(design_mat, b0, counts, n_dataset, masks, proportion, tolerance, max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    
    # Calculate standard error on predictors
    
    mat_w = np.diag(mfit)
    design_mat_t = design_mat.conj().transpose()
    try:
        vcov = np.linalg.inv(design_mat_t @ mat_w @ design_mat)
        predictors_sampled: npt.NDArray[Any] = np.random.multivariate_normal(
            beta, vcov, size=multinomial_size)
    except LinAlgError:
        warnings.warn("Could not determine confidence interval")
        return positive_estimate, positive_estimate, positive_estimate
    
    sampled_mfits = np.exp(design_mat @ predictors_sampled.T).T
    rounded_mfits = np.round(sampled_mfits)
    low_estimate = positive_estimate
    high_estimate = positive_estimate
    obs_pos = np.sum(counts[masks.positive_observed])
    set_num_threads(7)
    if True:
        results = rasch_parallel(design_mat, beta,  rounded_mfits, n_dataset, proportion, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                for (_, fitted, _) in results]
        low_estimate = np.percentile(estimates, 2.5)
        median_estimate = np.percentile(estimates, 50)
        high_estimate = np.percentile(estimates, 97.5)
    return median_estimate, low_estimate, high_estimate

def rasch_estimate_parametric_proportion(freq_df: pd.DataFrame,
                   n_dataset: int,
                   proportion: float,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
    #only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = rasch_em(design_mat, b0, counts, n_dataset, masks, proportion, tolerance, max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    if np.any(np.isnan(mfit)):
        estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
        stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
    else:
        p_vals = mfit / np.sum(mfit)
        multinomial_fits: npt.NDArray[Any] = np.random.multinomial(n_dataset, p_vals, multinomial_size)
        results = rasch_parallel(design_mat, b0, multinomial_fits, n_dataset, proportion, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                    for (_, fitted, _) in results]
        stats = EstimationModelStatistics(beta, mfit, deviance, (obs_pos + np.array(estimates)))
        low_estimate = obs_pos + np.percentile(estimates, 2.5)
        high_estimate = obs_pos + np.percentile(estimates, 97.5)
        estimate = Estimate(horizon_estimate, low_estimate, high_estimate)
    return estimate, stats

def rasch_estimate_parametric(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
    #only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    proportion: float = 100 / n_not_read
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = rasch_em(design_mat, b0, counts, n_dataset, masks, proportion, tolerance, max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    if np.any(np.isnan(mfit)):
        estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
        stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
    else:
        p_vals = mfit / np.sum(mfit)
        multinomial_fits: npt.NDArray[Any] = np.random.multinomial(n_dataset, p_vals, multinomial_size)
        rounded_beta = np.around(beta, 3)
        results = rasch_parallel(design_mat, b0, multinomial_fits, n_dataset, proportion, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                    for (_, fitted, _) in results]
        stats = EstimationModelStatistics(beta, mfit, deviance, (obs_pos + np.array(estimates)))
        low_estimate = obs_pos + np.percentile(estimates, 2.5)
        high_estimate = obs_pos + np.percentile(estimates, 97.5)
        estimate = Estimate(horizon_estimate, low_estimate, high_estimate)
    return estimate, stats

def rasch_estimate_parametric_no_fixed_proportion(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
   
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance, proportion = rasch_numpy(design_mat, counts, n_dataset, masks)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    if np.any(np.isnan(mfit)):
        estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
        stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
    else:
        p_vals = mfit / np.sum(mfit)
        multinomial_fits: npt.NDArray[Any] = np.random.multinomial(n_dataset, p_vals, multinomial_size)
        results = rasch_parallel(design_mat, b0, multinomial_fits, n_dataset, proportion, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                    for (_, fitted, _) in results]
        stats = EstimationModelStatistics(beta, mfit, deviance, (obs_pos + np.array(estimates)))
        low_estimate = obs_pos + np.percentile(estimates, 2.5)
        high_estimate = obs_pos + np.percentile(estimates, 97.5)
        estimate = Estimate(horizon_estimate, low_estimate, high_estimate)
    return estimate, stats

def rasch_estimate_parametric_init_by_pos(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000,
                   multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)
    
    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
   
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    rasch_pos_est, _ = rasch_estimate_only_pos(freq_df, 2000)
    proportion = (rasch_pos_est.point - obs_pos) / n_not_read
    beta, mfit, deviance = rasch_em(design_mat, b0, counts, n_dataset, masks, proportion)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    if np.any(np.isnan(mfit)):
        estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
        stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
    else:
        p_vals = mfit / np.sum(mfit)
        try:
            multinomial_fits: npt.NDArray[Any] = np.random.multinomial(n_dataset, p_vals, multinomial_size)
        except ValueError:
            estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
            stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
        else:
            results = rasch_parallel(design_mat, b0, multinomial_fits, n_dataset, proportion, masks)
            estimates: List[float] = [fitted[masks.positive_estimate][0]
                                        for (_, fitted, _) in results]
            stats = EstimationModelStatistics(beta, mfit, deviance, (obs_pos + np.array(estimates)))
            low_estimate = obs_pos + np.percentile(estimates, 2.5)
            high_estimate = obs_pos + np.percentile(estimates, 97.5)
            estimate = Estimate(horizon_estimate, low_estimate, high_estimate)
    return estimate, stats

def rasch_estimate(freq_df: pd.DataFrame,
                   n_dataset: int,
                   proportion: float,
                   tolerance: float = 1e-5,
                   max_it: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)

    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
    #only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = rasch_em(design_mat, b0, counts, n_dataset, masks, proportion, tolerance, max_it)
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    return Estimate(horizon_estimate, horizon_estimate, horizon_estimate), EstimationModelStatistics(beta, mfit, deviance, np.array([]))

def rasch_estimate_bf(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)

    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
    #only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = select_best_model(*rasch_bf(design_mat, b0, counts, n_dataset, masks))
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    return Estimate(horizon_estimate, horizon_estimate, horizon_estimate), EstimationModelStatistics(beta, mfit, deviance, np.array([]))

def rasch_estimate_bf_stats(freq_df: pd.DataFrame,
                   n_dataset: int,
                   proportion: float,
                   tolerance: float = 1e-5,
                   max_it: int = 2000):
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)

    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
    #only_pos_estimate = rasch_estimate_only_pos(freq_df)[0].point
    
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    betas, mfits, deviances = rasch_bf(design_mat, b0, counts, n_dataset, masks)
    ests = mfits[:,masks.positive_estimate]
    return ests, deviances 

def rasch_estimate_bf_parametric(freq_df: pd.DataFrame,
                   n_dataset: int,
                   tolerance: float = 1e-5,
                   max_it: int = 2000, 
                   multinomial_size: int = 2000) -> Tuple[Estimate, EstimationModelStatistics]:
    # Calculate general frequency statistics
    counts: npt.NDArray[Any] = freq_df["count"].values  # type: ignore
    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))

    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
                        .filter(regex=r"^(?:(?!positive$).)+$")
                        .filter(regex=r"^learner")
                        )
    df_formatted = l2_format(df_w_missing, learner_cols)

    # Calculate row masks to select data from the matrices and vectors
    est_mask: List[bool] = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask: List[bool] = list(df_formatted.positive > 0)
    masks = create_mask(pos_mask, est_mask)

    design_mat: npt.NDArray[Any] = (
        df_formatted.loc[:, df_formatted.columns != 'count']  # type: ignore
        .values)  # type: ignore

    counts: npt.NDArray[Any] = df_formatted["count"].values  # type: ignore
    obs_pos = np.sum(counts[masks.positive_observed])
        
    n_not_read = n_dataset - np.sum(counts[masks.observed])
    b0 = initial_fit(n_not_read, design_mat)
    beta, mfit, deviance = select_best_model(*rasch_bf(design_mat, b0, counts, n_dataset, masks))
    positive_estimate: float = mfit[masks.positive_estimate][0]
    horizon_estimate = obs_pos + positive_estimate
    proportion = positive_estimate / n_not_read
    if np.any(np.isnan(mfit)):
        estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
        stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
    else:
        p_vals = mfit / np.sum(mfit)
        try:
            multinomial_fits: npt.NDArray[Any] = np.random.multinomial(n_dataset, p_vals, multinomial_size)
        except ValueError:
            estimate = Estimate(horizon_estimate, horizon_estimate, horizon_estimate)
            stats = EstimationModelStatistics(beta, mfit, deviance, np.array([]))
            return estimate, stats
        results = rasch_parallel(design_mat, beta, multinomial_fits, n_dataset, proportion, masks)
        estimates: List[float] = [fitted[masks.positive_estimate][0]
                                    for (_, fitted, _) in results]
        stats = EstimationModelStatistics(beta, mfit, deviance, (obs_pos + np.array(estimates)))
        low_estimate = obs_pos + np.percentile(estimates, 2.5)
        high_estimate = obs_pos + np.percentile(estimates, 97.5)
        estimate = Estimate(horizon_estimate, low_estimate, high_estimate)
    return estimate, stats




class EMRaschRidgeParametricConvPython(
        EMRaschCombined[KT, DT, VT, RT, LT],
        Generic[KT, DT, VT, RT, LT]):
    def __init__(self):
        super().__init__()
        self.df: Optional[pd.DataFrame] = None
        self.est = float("nan")
        self.est_low = float("nan")
        self.est_high = float("nan")
        self.dfs = collections.deque()

    def _start_r(self) -> None:
        pass

    def calculate_estimate(self,
                           estimator: Estimator[Any, KT, DT, VT, RT, LT],
                           label: LT) -> Tuple[float, float, float]:
        pos_count = estimator.env.labels.document_count(label)
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if self.df is None or not self.df.equals(df):
            self.dfs.append(df)
            self.df = df
            self.est, self.est_low, self.est_up = rasch_estimate_parametric_approx(
                df, dataset_size)
        horizon = self.est + pos_count
        horizon_low = self.est_low + pos_count
        horizon_up = self.est_up + pos_count
        return horizon, horizon_low, horizon_up






class FastEMRaschPosNeg(
        EMRaschCombined[KT, DT, VT, RT, LT],
        Generic[KT, DT, VT, RT, LT]):

    estimates: Deque[Estimate]
    model_info: Deque[EstimationModelStatistics]
    dfs: Deque[pd.DataFrame]

    def __init__(self, multinomial_size: int = 2000):
        super().__init__()
        self.estimates = collections.deque()
        self.dfs = collections.deque()
        self.model_info = collections.deque()
        self.multinomial_size = multinomial_size

    def _start_r(self) -> None:
        pass

    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        assert isinstance(learner, Estimator)
        return self.calculate_estimate(learner, label)

    def calculate_estimate(self,
                           estimator: Estimator[Any, KT, DT, VT, RT, LT],
                           label: LT) -> Estimate:
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if not self.dfs or not self.dfs[-1].equals(df):
            self.dfs.append(df)
            est, stats = rasch_estimate_bf(df, dataset_size)
            self.estimates.append(est)
            self.model_info.append(stats)
        return self.estimates[-1]

class FastOnlyPos(FastEMRaschPosNeg):
    def calculate_estimate(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        df = self.get_occasion_history(estimator, label)
        if not self.dfs or not self.dfs[-1].equals(df):
            self.dfs.append(df)
            est, stats = rasch_estimate_only_pos(df, multinomial_size=self.multinomial_size)
            self.estimates.append(est)
            self.model_info.append(stats)
        return self.estimates[-1]

class FastPosAssisted(FastEMRaschPosNeg):
    def calculate_estimate(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT) -> Estimate:
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if not self.dfs or not self.dfs[-1].equals(df):
            self.dfs.append(df)
            est, stats = rasch_estimate_parametric_init_by_pos(df, dataset_size, multinomial_size=self.multinomial_size)
            self.estimates.append(est)
            self.model_info.append(stats)
        return self.estimates[-1]



from os import PathLike
from pathlib import Path
from typing import Any, Dict, FrozenSet, Generic, List, Sequence, Tuple

import numpy as np
import pandas as pd

from ..activelearning import ActiveLearner
from ..activelearning.estimator import Estimator
from ..typehints.typevars import DT, KT, LT, RT, VT
from ..utils.func import all_subsets, intersection, powerset, not_in_supersets, union

import functools
import string

def count_learners(df: pd.DataFrame) -> int:
    cols = df.filter(regex="learner").filter(regex="^((?!-positive).)*$").columns
    return len(cols.to_list())

def colnames(df: pd.DataFrame) -> Sequence[str]:
    table = df.filter(regex="learner|positive").filter(regex="^((?!-positive).)*$").values.tolist()
    options = string.ascii_uppercase[0:n_learners(df)] + "+"
    def match_chars(ar: List[int]) -> str:
        zipped = zip(ar, options)
        filtered = "".join([y for (x,y) in zipped if x == 1])
        if "+" not in filtered:
            return filtered + "-"
        return filtered
    colnames = list(map(match_chars, table))
    return colnames
    
def sorted_colnames(learners: int) -> Sequence[str]:
    learner_combos = sorted(
            ["".join(sorted(s)) for s in 
            powerset(string.ascii_uppercase[0:4]) if s],
            key=lambda item: (len(item), item))
    pos_cols = [f"{col}+" for col in learner_combos]
    neg_cols = [f"{col}-" for col in learner_combos]
    return pos_cols + neg_cols

def convert_df(df: pd.DataFrame):
    counts = np.array(df["count"].values.tolist()).reshape((1,-1)) # type: ignore
    columns = colnames(df)
    n_learners = count_learners(df)
    new_row = pd.DataFrame(data=counts, columns=columns)
    formatted = new_row[sorted_colnames(n_learners)]
    return formatted

# %%
class TableCollector(Generic[LT]):
    def __init__(self, pos_label: LT):
        self.dfs = list()
        self.pos_label = pos_label

    def get_contingency_sets(self, 
                         estimator: Estimator[Any, KT, Any, Any, Any, LT], 
                         label: LT) -> Dict[FrozenSet[int], FrozenSet[KT]]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(
                label).intersection(learner.env.labeled)
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: intersection(*[learner_sets[key] for key in combination])
            for combination in key_combinations
            if len(combination) >= 1
        }
        filtered_result = not_in_supersets(result)
        return filtered_result
    
    def get_occasion_history(self, 
                             estimator: Estimator[Any, Any, Any, Any, Any, LT], 
                             label: LT) -> pd.DataFrame:
        contingency_sets_pos = self.get_contingency_sets(estimator, label)
        contingency_sets_neg = self.get_contingency_sets_negative(estimator, label)
        learner_keys = union(*contingency_sets_pos.keys())
        
        rasch_pos_func = functools.partial(
            self.rasch_row, all_learners=learner_keys, positive = True)
        rasch_neg_func = functools.partial(
            self.rasch_row, all_learners=learner_keys, positive = False)
        
        
        rows_pos = list(map(rasch_pos_func, contingency_sets_pos.items()))
        rows_neg = list(map(rasch_neg_func, contingency_sets_neg.items()))

        rows = dict(enumerate(rows_pos + rows_neg))
        
        df = pd.DataFrame.from_dict(# type: ignore
            rows, orient="index")
        return df

    def get_contingency_sets_negative(self, 
                                      estimator: Estimator[Any, KT, Any, Any, Any, LT], 
                                      label: LT) -> Dict[FrozenSet[int], FrozenSet[KT]]:
        learner_sets = {
                learner_key: (
                    frozenset(learner.env.labeled).
                        difference(
                            learner.env.labels.get_instances_by_label(label))
                )
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: intersection(*[learner_sets[key] for key in combination])
            for combination in key_combinations
            if len(combination) >= 1
        }
        filtered_result = not_in_supersets(result)
        return filtered_result

    @staticmethod
    def rasch_row(combination: Tuple[FrozenSet[int], FrozenSet[Any]], 
              all_learners: FrozenSet[int],
              positive: bool) -> Dict[str, int]:
        learner_set, instances = combination
        learner_cols = {
            f"learner_{learner_key}": int(learner_key in learner_set) 
            for learner_key in all_learners
        }
        count_col = {"count": len(instances)}
        positive_col = {"positive": int(positive)}
        interaction_cols = {
            f"h{i-1}": len(all_subsets(learner_set, i, i)) 
            for i in range(2, len(all_learners))
        }
        pos_learner_cols = {
            f"learner_{learner_key}-positive": (
                int(learner_key in learner_set) if positive else 0)
            for learner_key in all_learners
        }
        interaction_pos_cols = {
            f"h{i-1}-positive": (
                len(all_subsets(learner_set, i, i)) if positive else 0)
            for i in range(2, len(all_learners))
        }
        final_row = {
            **learner_cols,
            **positive_col,
            **pos_learner_cols,
            **interaction_cols,
            **interaction_pos_cols,
            **count_col
        }
        return final_row

    def __call__(self, learner: ActiveLearner):
        if isinstance(learner, Estimator):
            df = self.get_occasion_history(learner, self.pos_label)
            self.dfs.append(df)

    @property
    def compact(self) -> pd.DataFrame:
        compact = pd.concat(map(convert_df, self.dfs), ignore_index=True)
        return compact # type: ignore

    def save_to_folder(self, path: "PathLike[str]") -> None:
        path = Path(path)
        if not path.exists():
            path.mkdir(parents=True)
        compact = self.compact
        compact.to_csv((path / "aggregated.csv"), index=False)
        for i, df in enumerate(self.dfs):
            df.to_csv((path / f"design_matrix_{i}.csv"), index=False)


from __future__ import annotations
from allib.utils.func import all_subsets, intersection, list_unzip3, union
from allib.activelearning.base import ActiveLearner

import collections
import itertools
import logging
import os
from abc import ABC, abstractmethod
from typing import (Any, Deque, Dict, FrozenSet, Generic, Iterable, List,
                    Optional, Sequence, Tuple, TypeVar)

import numpy as np
import numpy.typing as npt
import pandas as pd  # type: ignore

from ..activelearning.estimator import Estimator

from .base import AbstractEstimator
from ..utils.func import powerset, not_in_supersets

try:
    import rpy2.robjects as ro  # type: ignore
    from rpy2.robjects import pandas2ri  # type: ignore
    from rpy2.robjects.conversion import localconverter  # type: ignore
    from rpy2.robjects.packages import importr  # type: ignore
except ImportError:
    R_AVAILABLE = False
else:
    R_AVAILABLE = True

DT = TypeVar("DT")
VT = TypeVar("VT")
KT = TypeVar("KT")
LT = TypeVar("LT")
RT = TypeVar("RT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")
_T = TypeVar("_T")
_U = TypeVar("_U")

LOGGER = logging.getLogger(__name__)

def _check_R():
    """Checks if Python <-> R interop is available

    Raises
    ------
    ImportError
        If the interop is not available
    """    
    if not R_AVAILABLE:
        raise ImportError("Install rpy2 interop")

class AbundanceEstimator(AbstractEstimator, Generic[KT, DT, VT, RT, LT]):
    name = "BestBICEstimator"
    def __init__(self):
        self.matrix_history: Deque[pd.DataFrame] = collections.deque()
        self.contingency_history: Deque[Dict[FrozenSet[int], int]] = collections.deque()
        self._start_r()

    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "multiple_estimate.R")
        R["source"](r_script_file)
       
    def get_label_matrix(self, 
                         estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                         label: LT) -> pd.DataFrame:
        rows = {ins_key: {
            l_key: ins_key in learner.env.labeled
            for l_key, learner in enumerate(estimator.learners)}
            for ins_key in estimator.env.labels.get_instances_by_label(label)
        }
        dataframe = pd.DataFrame.from_dict(  # type: ignore
            rows, orient="index")
        self.matrix_history.append(dataframe)
        return dataframe

    def get_contingency_list(self, 
                         estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                         label: LT) -> Dict[FrozenSet[int], int]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(
                label).intersection(learner.env.labeled)
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: len(intersection(
                *[learner_sets[key] for key in combination]))
            for combination in key_combinations
            if len(combination) >= 1
        }
        self.contingency_history.append(result)
        return result

    def get_matrix(self, 
                   estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                   label: LT) -> npt.NDArray[Any]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(
                label).intersection(learner.env.labeled)
            for learner_key, learner in enumerate(estimator.learners)
        }
        n_learners = len(learner_sets)
        matrix = np.zeros(shape=(n_learners, n_learners))
        for i, key_a in enumerate(learner_sets):
            instances_a = learner_sets[key_a]
            for j, key_b in enumerate(learner_sets):
                if i != j:
                    instances_b = learner_sets[key_b]
                    intersection = instances_a.intersection(instances_b)
                    matrix[i, j] = len(intersection)
        return matrix

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df = self.get_label_matrix(estimator, label)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(df)
            abundance_r = ro.globalenv["get_abundance"]
            r_df = abundance_r(df_r)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    def calculate_abundance(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        ok_fit = res_df[res_df.infoFit == 0]
        if len(ok_fit) == 0:
            ok_fit = res_df
        best_result = ok_fit[ok_fit.BIC == ok_fit.BIC.min()]
        best_result = best_result[["abundance", "stderr"]]
        best_np = best_result.values
        return best_np[0, 0], best_np[0, 1]

    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        abundance, error = self.calculate_abundance(learner, label)
        return abundance, abundance-error, abundance + error

    def all_estimations(self, 
                        estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                        label: LT) -> Sequence[Tuple[str, float, float]]:
        res_df = self.calculate_abundance_R(estimator, label)
        ok_fit = res_df[res_df.infoFit == 0]
        if len(ok_fit) == 0:
            ok_fit = res_df
        results = ok_fit.values
        names = list(ok_fit.index)
        estimations = list(results[:,0])
        errors = list(results[:,1])
        tuples = list(zip(names, estimations, errors))
        return tuples
    
    def get_contingency_sets(self, 
                         estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                         label: LT) -> Dict[FrozenSet[int], FrozenSet[KT]]:
        learner_sets = {
            learner_key: learner.env.labels.get_instances_by_label(
                label).intersection(learner.env.labeled)
            for learner_key, learner in enumerate(estimator.learners)
        }
        key_combinations = powerset(range(len(estimator.learners)))
        result = {
            combination: intersection(*[learner_sets[key] for key in combination])
            for combination in key_combinations
            if len(combination) >= 1
        }
        filtered_result = not_in_supersets(result)
        return filtered_result

    def get_occasion_history(self, 
                             estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                             label: LT) -> pd.DataFrame:
        contingency_sets = self.get_contingency_sets(estimator, label)
        learner_keys = union(*contingency_sets.keys())
        rows = {i:
            {
                **{
                    f"learner_{learner_key}": int(learner_key in combination) 
                    for learner_key in learner_keys
                },  
                **{
                    "count": len(instances)
                }
            }
            for (i, (combination, instances)) in enumerate(contingency_sets.items())
        }
        df = pd.DataFrame.from_dict(# type: ignore
            rows, orient="index")
        return df

class MeanAbundanceEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "MeanAbundance"
    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        _, estimations, errors = list_unzip3(self.all_estimations(learner, label))
        estimation = float(np.mean(estimations)) #type: ignore
        error = float(np.mean(errors)) # type: ignore
        return estimation, estimation - error, estimation + error

class WeightedMeanAbundanceEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "WeightedBICEstimator"
    def calculate_abundance(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        ok_fit = res_df[res_df.infoFit == 0]
        if len(ok_fit) == 0:
            ok_fit = res_df
        
        best_bic = ok_fit.BIC.min()
        worst_bic = ok_fit.BIC.max()
        scaled_bics = ok_fit.BIC.max() / ok_fit.BIC
        total_bic_parts = sum(scaled_bics)
        weights = scaled_bics / total_bic_parts

        weighted_abundance = sum(weights * ok_fit.abundance)
        weighted_stderr = sum(weights * ok_fit.stderr)
        return weighted_abundance, weighted_stderr

class MedianAbundanceEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "MedianAbundanceEstimator"
    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        _, estimations, errors = list_unzip3(self.all_estimations(learner, label))
        estimation = float(np.median(estimations)) # type: ignore
        error = float(np.median(errors)) # type: ignore
        return estimation, estimation - error, estimation + error

class NegativeAbundanceEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    def __init__(self, neg_label: LT):
        super().__init__()
        self.neg_label = neg_label
    
    def __call__(self, learner: ActiveLearner[Any, KT, DT, VT, RT, LT], label: LT) -> Tuple[float, float, float]:
        if not isinstance(learner, Estimator):
            return 0.0, 0.0, 0.0
        neg_docs = learner.env.labels.document_count(self.neg_label)
        neg_abundance, error = self.calculate_abundance(learner, self.neg_label)
        pos_docs = learner.env.labels.document_count(label)
        estimate = pos_docs * (1 + (neg_docs /neg_abundance))
        error_estimate = (neg_docs / error) * pos_docs
        return estimate, estimate - error_estimate, estimate + error_estimate



class RaschEstimator(AbundanceEstimator[KT, DT, VT, RT, LT], Generic[KT, DT, VT, RT, LT]):
    name = "RaschEstimator"
    def _start_r(self) -> None:
        _check_R()
        R = ro.r
        filedir = os.path.dirname(os.path.realpath(__file__))
        r_script_file = os.path.join(filedir, "rasch_estimate.R")
        R["source"](r_script_file)

    def get_occasion_history(self, 
                             estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                             label: LT) -> pd.DataFrame:
        contingency_sets = self.get_contingency_sets(estimator, label)
        learner_keys = union(*contingency_sets.keys())
        max_len = len(learner_keys)
        rows = {i:
            {
                **{
                    f"learner_{learner_key}": int(learner_key in combination) 
                    for learner_key in learner_keys
                },  
                **{
                    "count": len(instances),
                    "h1": len(all_subsets(combination, 2, max_len - 1))
                }
            }
            for (i, (combination, instances)) in enumerate(contingency_sets.items())
        }
        df = pd.DataFrame.from_dict(# type: ignore
            rows, orient="index")
        self.matrix_history.append(df)
        return df

    def calculate_abundance_R(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                              label: LT) -> pd.DataFrame:
        df = self.get_occasion_history(estimator, label)
        with localconverter(ro.default_converter + pandas2ri.converter):
            df_r = ro.conversion.py2rpy(df)
            abundance_r = ro.globalenv["get_rasch"]
            r_df = abundance_r(df_r)
            res_df = ro.conversion.rpy2py(r_df)
        return res_df

    
    def calculate_abundance(self, 
                            estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                            label: LT) -> Tuple[float, float]:
        res_df = self.calculate_abundance_R(estimator, label)
        estimate_missing = res_df.values[0,0]
        estimate_error = res_df.values[0,1]
        total_found = estimator.env.labels.document_count(label)
        return (total_found + estimate_missing), estimate_error * 2

    def all_estimations(self, estimator: Estimator[Any, KT, DT, VT, RT, LT], label: LT) -> Sequence[Tuple[str, float, float]]:
        return []
    
    


from typing import TypeVar, Union, Any

from instancelib.instances import Instance

KT = TypeVar("KT")

def to_key(instance_or_key: Union[KT, Instance[KT, Any, Any, Any]]) -> KT:
    """Returns the identifier of the instance if `instance_or_key` is an `Instance`
    or return the key if `instance_or_key` is a `KT`

    Parameters
    ----------
    instance_or_key : Union[KT, Instance]
        An implementation of `Instance` or an identifier typed variable

    Returns
    -------
    KT
        The identifer of the instance (or the input verbatim)
    """
    if isinstance(instance_or_key, Instance):
        key: KT = instance_or_key.identifier # type: ignore
        return key
    return instance_or_key


from typing import Any, Dict

from lightgbm.sklearn import LGBMClassifier  # type: ignore

from sklearn.ensemble import RandomForestClassifier  # type: ignore
from sklearn.linear_model import LogisticRegression  # type: ignore
from sklearn.multiclass import OneVsRestClassifier  # type: ignore
from sklearn.multioutput import ClassifierChain  # type: ignore
from sklearn.naive_bayes import MultinomialNB  # type: ignore
from sklearn.pipeline import Pipeline  # type: ignore
from sklearn.preprocessing import (LabelBinarizer,  # type: ignore
                                   MultiLabelBinarizer)
from sklearn.svm import SVC, LinearSVC  # type: ignore

from ..balancing import BalancerFactory
from ..balancing.catalog import BalancerCatalog as BL
from ..factory import AbstractBuilder, ObjectFactory
from ..machinelearning import MultilabelSkLearnClassifier, SkLearnClassifier
from ..module.component import Component
from .catalog import MachineLearningCatalog as ML


class ClassifierBuilder(AbstractBuilder):
    def __call__(self, task, **kwargs):
        return self._factory.create(task, **kwargs)

class BinaryClassificationBuilder(AbstractBuilder):
    def __call__(self,  # type: ignore
            sklearn_model: ML.SklearnModel, 
            model_configuration: Dict,
            balancer: Dict[str, Any],  
            storage_location = None, filename=None, **kwargs):
        encoder = LabelBinarizer()
        built_balancer = self._factory.create(Component.BALANCER, **balancer)
        classifier = self._factory.create(sklearn_model, **model_configuration)
        return SkLearnClassifier(
            classifier, encoder, built_balancer, storage_location=storage_location, filename=filename)

class SklearnBuilder(AbstractBuilder):
    def __call__(self, sk_type: ML.SklearnModel, sklearn_config, **kwargs): # type: ignore
        return self._factory.create(sk_type, **sklearn_config)

class MulticlassBuilder(AbstractBuilder):
    def __call__(self, mc_method: ML.MulticlassMethod, **kwargs): # type: ignore
        return self._factory.create(mc_method, **kwargs)

class MultilabelBuilder(AbstractBuilder):
    def __call__(self, mc_method: ML.MulticlassMethod, **kwargs): # type: ignore
        encoder = MultiLabelBinarizer()
        classifier = self._factory.create(mc_method, **kwargs)
        return MultilabelSkLearnClassifier(classifier, encoder)

class OneVsRestBuilder(AbstractBuilder):
    def __call__(self, sklearn_model: ML.SklearnModel, model_configuration, **kwargs): # type: ignore
        base_classifier = self._factory.create(sklearn_model, **model_configuration)
        return OneVsRestClassifier(base_classifier)

class MachineLearningFactory(ObjectFactory):
    def __init__(self) -> None:
        super().__init__()
        self.attach(BalancerFactory())
        self.register_builder(Component.CLASSIFIER, ClassifierBuilder())
        self.register_builder(ML.Task.BINARY, BinaryClassificationBuilder())
        self.register_builder(ML.Task.MULTICLASS, MulticlassBuilder())
        self.register_builder(ML.Task.MULTILABEL, MultilabelBuilder())
        self.register_builder(ML.MulticlassMethod.ONE_VS_REST, OneVsRestBuilder())
        self.register_constructor(ML.SklearnModel.RANDOM_FOREST, RandomForestClassifier)
        self.register_constructor(ML.SklearnModel.NAIVE_BAYES, MultinomialNB)
        self.register_constructor(ML.SklearnModel.LOGISTIC, LogisticRegression)
        self.register_constructor(ML.SklearnModel.SVM, LinearSVC)
        self.register_constructor(ML.SklearnModel.SVC, SVC)
        self.register_constructor(ML.SklearnModel.LGBM, LGBMClassifier)



from __future__ import annotations

import functools
import os
import uuid
from abc import ABC, abstractmethod
from typing import (FrozenSet, Generic, Iterable, Iterator, List, Optional,
                    Sequence, Set, Tuple, TypeVar, Any)

from ..environment import AbstractEnvironment
from instancelib.instances.base import Instance

from ..typehints import KT, VT, LT, LVT, PVT


class AbstractClassifier(ABC, Generic[KT, VT, LT, LVT, PVT]):
    _name = "AbstractClassifier"

    @abstractmethod
    def __call__(self, 
            environment: AbstractEnvironment[Any, KT, Any, VT, Any, LT]
        ) -> AbstractClassifier[KT, VT, LT, LVT, PVT]:
        """Initialize the classifier by supplying the target labels
        
        Parameters
        ----------
        target_labels : Set[LT]
            A set with number of labels
        
        Returns
        -------
        AbstractClassifier
            [description]
        """        
        raise NotImplementedError

    @abstractmethod
    def fit(self, x_data: Sequence[VT], y_data: Sequence[LVT]):
        pass

    @abstractmethod
    def predict_proba(self, x_data: VT) -> PVT:
        pass

    @abstractmethod
    def predict(self, x_data: VT) -> LVT:
        pass

    @abstractmethod
    def encode_labels(self, labels: Iterable[LT]) -> LVT:
        pass

    @abstractmethod
    def predict_instances(self, instances: Sequence[Instance[KT, Any, VT, Any]]) -> Sequence[FrozenSet[LT]]:
        pass

    @abstractmethod
    def fit_vectors(self, x_data: Sequence[VT], labels: Sequence[FrozenSet[LT]]):
        pass

    @abstractmethod
    def fit_instances(self, instances: Sequence[Instance[KT, Any, VT, Any]], labels: Sequence[FrozenSet[LT]]):
        pass

    @abstractmethod
    def predict_proba_instances(self, instances: Sequence[Instance[KT, Any, VT, Any]]) -> Sequence[FrozenSet[Tuple[LT, float]]]:
        pass

    @property
    def name(self) -> str:
        return self._name
        
    @property
    @abstractmethod
    def fitted(self) -> bool:
        pass

    @abstractmethod
    def get_label_column_index(self, label: LT) -> int:
        raise NotImplementedError



from enum import Enum

class MachineLearningCatalog:
    class Task(Enum):
        SEQUENCE = "Sequence"
        BINARY = "Binary"
        MULTICLASS = "MultiClass"
        MULTILABEL = "MultiLabel"

    class MulticlassMethod(Enum):
        ONE_VS_REST = "OneVsRest"
        BINARY_RELEVANCE = "BinaryRelevance"

    class SklearnModel(Enum):
        RANDOM_FOREST = "RandomForest"
        NAIVE_BAYES = "NaiveBayes"
        SVM = "SVM"
        SVC = "SVC"
        LOGISTIC = "LogisticRegression"
        LGBM = "LGBMClassifier"


from ..utils.random import get_random_generator
from typing import Generic, Optional, TypeVar, Any

import numpy as np  # type: ignore
import numpy.typing as npt
from scipy.stats import entropy  # type: ignore

from .selectioncriterion import AbstractSelectionCriterion
from .labelmethods import LabelProbabilityBased
from .catalog import ALCatalog

DT = TypeVar("DT")
VT = TypeVar("VT")
KT = TypeVar("KT")
LT = TypeVar("LT")
RT = TypeVar("RT")
LVT = TypeVar("LVT")
PVT = TypeVar("PVT")


class LeastConfidence(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.LEAST_CONFIDENCE

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        max_prob = prob_mat.max(axis=1)  # type: ignore
        return 1 - max_prob  # type: ignore


class NearDecisionBoundary(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.NEAR_DECISION_BOUNDARY

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        min_prob = np.abs(prob_vec - 0.5).min(axis=1)  # type: ignore
        return -1.0 * min_prob  # type: ignore


class MarginSampling(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.MARGIN_SAMPLING

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        # select the two highest probabilities per instance
        # with the lowest of the two on [:,0] and the highest on [:,1]
        two_best_prob = np.sort(prob_mat, axis=1)[:, -2:]  # type: ignore
        # subtract
        margin = np.diff(two_best_prob, axis=1)
        # invert in order to use as a maximization method
        return -margin


class EntropySampling(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.MAX_ENTROPY

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        return entropy(prob_mat, axis=1)  # type: ignore


class LabelUncertainty(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.LABELUNCERTAINTY

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        min_prob = -np.abs(prob_mat - 0.5)
        return min_prob


class LabelUncertaintyNew(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.LABELUNCERTAINTY_NEW

    def __init__(self, label_column: int):
        super().__init__()
        self.label_column = label_column

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        prob_mat_sliced = prob_mat[:, self.label_column]
        min_prob = -np.abs(prob_mat_sliced - 0.5)
        return min_prob


class RandomMLStrategy(AbstractSelectionCriterion):
    name = ALCatalog.QueryType.RANDOM_ML

    def __init__(self, *_, rng: Optional[np.random.Generator] = None):
        self.rng = get_random_generator(rng)

    def __call__(self, prob_mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        prob_mat_size = prob_mat.shape[0]
        random_score = self.rng.random((prob_mat_size,))
        return random_score



import itertools
import warnings
from logging import currentframe
from typing import Any, Generic, List, Optional, Sequence, Tuple

import numpy as np
import numpy.typing as npt
import pandas as pd
from numpy.linalg import LinAlgError

from ..activelearning.estimator import Estimator
from ..typehints.typevars import DT, KT, LT, RT, VT
from .rasch import EMRaschCombined


def l2(b0: npt.NDArray[Any], 
       design_mat: npt.NDArray[Any], 
       counts: npt.NDArray[Any],
       N: int,
       lam: float = 0, 
       tolerance: float = 1e-8,
       max_it: int = 10000,
       epsilon: float = 0.1
       ) -> npt.NDArray[Any]:
    counts = counts + epsilon
    current_tolerance = 1
    b = b0
    it = 0
    while current_tolerance > tolerance and it < max_it:
        it = it + 1

        mu = np.exp(design_mat @ b0)
        all_but_first_b0: npt.NDArray[Any] = b0[1:]
        lbpart: npt.NDArray[Any] = 2 * np.concatenate(
            [
                np.array([0]), 
                (lam * all_but_first_b0)
            ])
        design_mat_t = design_mat.conj().transpose()
        db = design_mat_t @ counts  - design_mat_t @ mu - lbpart
        repeat_count = b0.shape[0] - 1
        other_lb_part = np.hstack([[0], np.repeat(lam, repeat_count)])
        I = (design_mat_t @ np.diag(mu)) @ design_mat + 2 * np.diag(other_lb_part)
        try:
            invI = np.linalg.inv(I)
        except LinAlgError:
            # Change lambda if the matrix is not invertible
            lam = 1e-6 if lam == 0 else 10 * lam
        else:
            # If inversion succeeded, calculate new coefficients b
            b = b0 + invI @ db
            current_tolerance = np.sum(np.abs(b0 - b))
            b0 = b
    if it >= max_it:
        warnings.warn(f"Exceeded maximum L2 iterations ({max_it}). Estimate might not be accurate")
    return b

def calc_deviance(y: npt.NDArray[Any], y_hat: npt.NDArray[Any]) -> float:
    pos_y_idx = np.where(y > 0)
    pos_y = y[pos_y_idx]
    deviance = (
        2 * np.sum(pos_y * np.log(pos_y / y_hat[pos_y_idx])) -
        2 * np.sum(y - y_hat))
    return deviance

def add_missing_count_rows(freq_df: pd.DataFrame, n_pos: float, n_neg: float) -> pd.DataFrame:
    pos_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_pos, "positive": 1}
    }
    neg_dict = {
        **{col: 0 for col in freq_df.columns},
        **{"count": n_neg, "positive": 0}
    }
    new_freq_df = (
        freq_df
            .append(pos_dict, ignore_index=True)
            .append(neg_dict, ignore_index=True)
    )   
    return new_freq_df

def l2_format(freq_df: pd.DataFrame, learner_cols: Sequence[str]) -> pd.DataFrame:
    order_cols = ["positive", *learner_cols]
    df = freq_df.sort_values(order_cols)
    df.insert(0, "intercept", 1)
    return df

def parameter_ci(estimate: float, se: float) -> Tuple[float, float]:
    var_min = estimate - 2 * se
    var_max = estimate + 2 * se
    return var_max, var_min

def calculate_estimate(intercept: float, pos: float) -> float:
    return np.exp(intercept + pos)

def calculate_ci_rasch(intercept: float, 
                       positive: float, 
                       intercept_se: float, 
                       positive_se: float) -> Tuple[float, float]:
    intercepts = parameter_ci(intercept, intercept_se)
    positive_par = parameter_ci(positive, positive_se)
    combinations = itertools.product(intercepts, positive_par)
    results = list(itertools.starmap(calculate_estimate, combinations))
    min_result = min(results)
    max_result = max(results)
    return min_result, max_result


def rasch_estimate(freq_df: pd.DataFrame, 
                n_dataset: int,
                proportion: float = 0.1,
                tolerance: float = 1e-5, max_it: int = 1000) -> Tuple[float, float, float]:
    # Calculate general frequency statistics
    n_notread = n_dataset - np.sum(freq_df["count"].values)
    n_pos = proportion * n_notread
    n_neg = (1 - proportion) * n_dataset

    # Add place holder rows for the counts that we aim to estimate
    df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))
    
    # Change the dataframe in the correct format for the L2 algorithm
    learner_cols = list(df_w_missing
        .filter(regex=r"^(?:(?!positive$).)+$")
        .filter(regex=r"^learner")
        )
    df_formatted = l2_format(df_w_missing, learner_cols)
    
    # Gather some column indices that we need below to access data from the matrices
    positive_col_idx = list(df_formatted.columns).index("positive")
    all_pos_cols = list(df_formatted
        .filter(regex=r"^positive$"))

    # Calculate row masks to select data from the matrices and vectors
    obs_mask = list(df_formatted[learner_cols].sum(axis=1) > 0)
    est_mask = list(df_formatted[learner_cols].sum(axis=1) <= 0)
    pos_mask = list(df_formatted.positive > 0)
    neg_mask = list(df_formatted.positive <= 0)
    neg_est_mask: List[bool] = [est & neg for est, neg in zip(est_mask, neg_mask)]
    pos_est_mask: List[bool] = [est & pos for est, pos in zip(est_mask, pos_mask)]

    design_mat = (
        df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
            .values) # type: ignore
    
    
    obs_counts = df_formatted["count"].values
    efit = df_formatted["count"].values
    efit[pos_est_mask] = n_pos # type: ignore
    efit[neg_est_mask] = n_neg # type: ignore
    mfit = efit

    # Calculate initial fit
    b0 = np.hstack(
        [
            np.log([n_dataset]), 
            np.repeat(0,design_mat.shape[1] - 1)
        ])
    beta = l2(b0, design_mat, efit, n_dataset)
    deviance = calc_deviance(
            obs_counts, 
            np.exp(np.matmul(design_mat, beta)))
    
    current_tolerance = 1
    
    # Expectation Maximization 
    it = 0
    while current_tolerance > tolerance:
        old_deviance = deviance
        mfit = np.exp(design_mat @ beta)
        efit[est_mask] = mfit[est_mask] * n_notread / np.sum(mfit[est_mask])
        beta = l2(beta, design_mat, efit, n_dataset)
        deviance = calc_deviance(
            obs_counts, 
            np.exp(design_mat @ beta))
        current_tolerance = old_deviance - deviance
        if it > max_it:
            warnings.warn(f"Exceeded maximum EM iterations ({max_it}). Estimate might not be accurate")
            return float("nan"), float("nan"), float("nan")


    # Calculate final results
    fitted = np.exp(design_mat @ beta)
    positive_estimate: float = fitted[pos_est_mask][0]

    # Calculate standard error on predictors
    mat_w = np.diag(fitted)
    design_mat_t = design_mat.conj().transpose()
    vcov = np.linalg.inv(design_mat_t @ mat_w @ design_mat)
    se = np.sqrt(np.diagonal(vcov))

    # Calculate confidence interval on estimates    
    lower_bound, upper_bound = calculate_ci_rasch(
        beta[0], 
        beta[positive_col_idx], 
        se[0], 
        se[positive_col_idx])
    
    return positive_estimate, lower_bound, upper_bound

class EMRaschRidgePython(
            EMRaschCombined[KT, DT, VT, RT, LT], 
            Generic[KT, DT, VT, RT, LT]):
    def __init__(self):
        super().__init__()
        self.df: Optional[pd.DataFrame] = None
        self.est = float("nan")
        self.est_low = float("nan")
        self.est_high = float("nan")
    
    def _start_r(self) -> None:
        pass
          
    def calculate_estimate(self, 
                           estimator: Estimator[Any, KT, DT, VT, RT, LT], 
                           label: LT) -> Tuple[float, float, float]:
        pos_count = estimator.env.labels.document_count(label)
        dataset_size = len(estimator.env.dataset)
        df = self.get_occasion_history(estimator, label)
        if self.df is None or not self.df.equals(df):
            self.df = df        
            self.est, self.est_low, self.est_up = rasch_estimate(df, dataset_size)
        horizon = self.est + pos_count
        horizon_low = self.est_low + pos_count
        horizon_up = self.est_up + pos_count
        return horizon, horizon_low, horizon_up

# def rasch_estimate_delta_method(freq_df: pd.DataFrame, 
#                 n_dataset: int,
#                 proportion: float = 0.1,
#                 tolerance: float = 1e-5) -> Tuple[float, float, float]:
#     # Calculate general frequency statistics
#     n_notread = n_dataset - np.sum(freq_df["count"].values)
#     n_pos = proportion * n_notread
#     n_neg = (1 - proportion) * n_dataset

#     # Add place holder rows for the counts that we aim to estimate
#     df_w_missing = add_missing_count_rows(freq_df, float("nan"), float("nan"))
    
#     # Change the dataframe in the correct format for the L2 algorithm
#     learner_cols = list(df_w_missing
#         .filter(regex=r"^(?:(?!positive$).)+$")
#         .filter(regex=r"^learner")
#         )
#     df_formatted = l2_format(df_w_missing, learner_cols)
    
#     # Gather some column indices that we need below to access data from the matrices
#     positive_col_idx = list(df_formatted.columns).index("positive")
#     all_pos_cols = list(df_formatted
#         .filter(regex=r"positive"))

#     # Calculate row masks to select data from the matrices and vectors
#     obs_mask = list(df_formatted[learner_cols].sum(axis=1) > 0)
#     est_mask = list(df_formatted[learner_cols].sum(axis=1) <= 0)
#     pos_mask = list(df_formatted.positive > 0)
#     neg_mask = list(df_formatted.positive <= 0)
#     neg_est_mask = [est & neg for est, neg in zip(est_mask, neg_mask)]
#     pos_est_mask = [est & pos for est, pos in zip(est_mask, pos_mask)]

#     design_mat = (
#         df_formatted.loc[:, df_formatted.columns != 'count'] # type: ignore
#             .values) # type: ignore
    
#     # Calculate start values
#     obs_counts = df_formatted["count"].values
#     efit = df_formatted["count"].values
#     efit[pos_est_mask] = n_pos
#     efit[neg_est_mask] = n_neg
#     mfit = efit

#     # Calculate initial fit
#     b0 = np.hstack(
#         [
#             np.log([n_dataset]), 
#             np.repeat(0,design_mat.shape[1] - 1)
#         ])
#     beta = l2(b0, design_mat, efit, n_dataset)
#     deviance = calc_deviance(
#             obs_counts, 
#             np.exp(np.matmul(design_mat, beta)))
    
#     current_tolerance = 1
    
#     # Expectation Maximization 
#     while current_tolerance > tolerance:
#         old_deviance = deviance
#         mfit = np.exp(design_mat @ beta)
#         efit[est_mask] = mfit[est_mask] * n_notread / np.sum(mfit[est_mask])
#         beta = l2(beta, design_mat, efit, n_dataset)
#         deviance = calc_deviance(
#             obs_counts, 
#             np.exp(design_mat @ beta))
#         current_tolerance = old_deviance - deviance


#     # Calculate final results
#     fitted = np.exp(design_mat @ beta)
#     positive_estimate: float = fitted[pos_est_mask][0]

#     # Calculate standard error on predictors using the delta method
#     mat_w = np.diag(fitted)
#     design_mat_t = design_mat.conj().transpose()
#     vcov = np.linalg.inv(design_mat_t @ mat_w @ design_mat)

    
#     design_mat_pos = np.vstack([
#         design_mat[pos_mask], design_mat[pos_mask]
#     ])
#     design_mat_neg = np.vstack([
#         design_mat[neg_mask], design_mat[neg_mask]
#     ])
#     pred_pos = design_mat_pos @ beta
#     pred_neg = design_mat_neg @ beta

#     deriv_pos = np.exp(pred_pos)
#     deriv_neg = np.exp(pred_neg)

#     jacobian_pos = deriv_pos @ design_mat_pos / design_mat_pos.shape[0]
#     jacobian_neg = deriv_neg @ design_mat_neg / design_mat_neg.shape[0]

#     variance_pos = jacobian_pos @ vcov @ jacobian_pos.transpose()
#     variance_neg = jacobian_neg @ vcov @ jacobian_neg.transpose()
    
#     se_pos = np.sqrt(variance_pos)
    

#     # Calculate confidence interval on estimates    
#     lower_bound = positive_estimate - 2 * se_pos
#     upper_bound = positive_estimate + 2 * se_pos
    
#     return positive_estimate, lower_bound, upper_bound



from typing import Tuple, Any

import imblearn # type: ignore
import numpy as np
import numpy.typing as npt
from .base import BaseBalancer

class RandomOverSampler(BaseBalancer):
    """Balancing class that undersamples the data with a given ratio.
    """
    name = "RandomOverSampler"

    def __init__(self, random_state=0):
        super().__init__()
        self.ros = imblearn.over_sampling.RandomOverSampler(random_state=random_state)

    def resample(self, x_data: npt.NDArray[Any], y_data: npt.NDArray[Any]) -> Tuple[npt.NDArray[Any], npt.NDArray[Any]]:
        x_resampled, y_resampled = self.ros.fit_resample(x_data, y_data) # type: ignore
        return x_resampled, y_resampled # type: ignore


from __future__ import annotations

import collections
import itertools
import logging
import math
import os
import random
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

from instancelib.typehints import DT, KT, LT, RT, VT

from ..activelearning.ensembles import ManualEnsemble
from ..activelearning.ml_based import MLBased
from ..environment.base import IT, AbstractEnvironment
from ..machinelearning import AbstractClassifier
from ..utils import get_random_generator
from .base import ActiveLearner

_T = TypeVar("_T")

LOGGER = logging.getLogger(__name__)


def intersection(first: FrozenSet[_T], *others: FrozenSet[_T]) -> FrozenSet[_T]:
    return first.intersection(*others)


def powerset(iterable: Iterable[_T]) -> FrozenSet[FrozenSet[_T]]:
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    result = itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )
    return frozenset(map(frozenset, result))  # type: ignore


class Estimator(
    ManualEnsemble[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        learners: Sequence[
            ActiveLearner[IT, KT, DT, VT, RT, LT],
        ],
        probabilities: Optional[Sequence[float]] = None,
        rng: Any = None,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        probs = (
            [1.0 / len(learners)] * len(learners)
            if probabilities is None
            else probabilities
        )
        super().__init__(env, learners, probs, rng, *_, identifier=identifier, **__)

    @classmethod
    def builder(
        cls,
        learner_builders: Sequence[
            Callable[..., ActiveLearner[IT, KT, DT, VT, RT, LT]]
        ],
        probabilities: Optional[Sequence[float]] = None,
        *_: Any,
        **__: Any,
    ) -> Callable[..., ManualEnsemble[IT, KT, DT, VT, RT, LT]]:
        probs = (
            [1.0 / len(learner_builders)] * len(learner_builders)
            if probabilities is None
            else probabilities
        )
        return super().builder(learner_builders, probs, *_, **__)


class RetryEstimator(
    Estimator[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    def __next__(self) -> IT:
        # Choose the next random learners
        learner = self._choose_learner()

        # Select the next instance from the learner
        ins = next(learner)

        # Check if the instance identifier has not been labeled already
        while ins.identifier in self.env.labeled:
            # This instance has already been labeled my another learner.
            # Skip it and mark as labeled
            learner.set_as_labeled(ins)
            LOGGER.info(
                "The document with key %s was already labeled. Skipping", ins.identifier
            )
            ins = next(learner)

        # Set the instances as sampled by learner with key al_idx and return the instance
        self._sample_dict[ins.identifier] = self.learners.index(learner)
        return ins


class CycleEstimator(
    Estimator[IT, KT, DT, VT, RT, LT], Generic[IT, KT, DT, VT, RT, LT]
):
    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, VT, RT, LT],
        learners: Sequence[ActiveLearner[IT, KT, DT, VT, RT, LT]],
        probabilities: Optional[Sequence[float]] = None,
        rng: Any = None,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> None:
        super().__init__(
            env, learners, probabilities, rng, *_, identifier=identifier, **__
        )
        self.learnercycle = itertools.cycle(self.learners)

    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, VT, RT, LT]:
        return next(self.learnercycle)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
)

import numpy.typing as npt

from ..environment.base import AbstractEnvironment
from ..machinelearning.base import AbstractClassifier
from ..typehints import DT, IT, KT, LT, RT
from .base import ActiveLearner
from .ensembles import AbstractEnsemble
from .selectioncriterion import AbstractSelectionCriterion
from .ml_based import FeatureMatrix, MLBased, ProbabilityBased
from .random import RandomSampling


class LabelProbabilityBased(
    ProbabilityBased[IT, KT, DT, RT, LT], ABC, Generic[IT, KT, DT, RT, LT]
):
    def __init__(
        self,
        env: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT],
        classifier: AbstractClassifier[
            KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        selection_criterion: AbstractSelectionCriterion,
        label: LT,
        fallback: ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT],
        identifier: Optional[str] = None,
        *_,
        **__,
    ) -> None:
        super().__init__(
            env, classifier, selection_criterion, fallback, identifier=identifier
        )
        self.label = label
        self.labelposition: int = self.classifier.get_label_column_index(self.label)

    @property
    def name(self) -> Tuple[str, LT]:
        if self.identifier is not None:
            return f"{self.identifier}", self.label
        return f"{self._name} :: {self.classifier.name}", self.label

    def _get_predictions(
        self, matrix: FeatureMatrix[KT]
    ) -> Tuple[Sequence[KT], npt.NDArray[Any]]:
        prob_vec: npt.NDArray[Any] = self.classifier.predict_proba(
            matrix.matrix
        )  # type: ignore
        # type: ignore
        sliced_prob_vec: npt.NDArray[Any] = prob_vec[:, self.labelposition]  # type: ignore
        keys = matrix.indices
        return keys, sliced_prob_vec

    @classmethod
    def builder(
        cls,
        classifier: AbstractClassifier[
            KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        selection_criterion: AbstractSelectionCriterion,
        fallback_builder: Callable[
            ..., ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]
        ],
        batch_size=200,
        *_,
        identifier: Optional[str] = None,
        **__,
    ) -> Callable[..., ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT],]:
        def wrap_func(
            env: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT], 
            pos_label: LT, *_, **__
        ):
            built_classifier = classifier(env)
            fallback = fallback_builder(env)
            return cls(
                env,
                built_classifier,
                selection_criterion,
                pos_label,
                fallback,
                batch_size=batch_size,
                identifier=identifier,
            )
        return wrap_func


class LabelEnsemble(
    AbstractEnsemble[IT, KT, DT, npt.NDArray[Any], RT, LT],
    MLBased[IT, KT, DT, npt.NDArray[Any], RT, LT, npt.NDArray[Any], npt.NDArray[Any]],
    Generic[IT, KT, DT, RT, LT],
):
    _name = "LabelEnsemble"

    def __init__(
        self,
        classifier: AbstractClassifier[
            KT, npt.NDArray[Any], LT, npt.NDArray[Any], npt.NDArray[Any]
        ],
        al_method: Callable[..., LabelProbabilityBased[IT, KT, DT, RT, LT]],
        *_,
        **__,
    ) -> None:
        self._al_builder = al_method
        self.label_dict: Dict[LT, int] = dict()
        self.learners: List[
            ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]
        ] = list()
        self.classifier = classifier
        self._has_ordering = False

    def __call__(
        self, environment: AbstractEnvironment[IT, KT, DT, npt.NDArray[Any], RT, LT]
    ) -> LabelEnsemble[IT, KT, DT, RT, LT]:
        super().__call__(environment)
        labelset = self.env.labels.labelset
        self.label_dict = {label: idx for idx, label in enumerate(labelset)}
        self.learners = [
            self._al_builder(self.classifier, label)(environment) for label in labelset
        ]
        return self

    def _choose_learner(self) -> ActiveLearner[IT, KT, DT, npt.NDArray[Any], RT, LT]:
        labelcounts = [
            (self.env.labels.document_count(label), label)
            for label in self.env.labels.labelset
        ]
        min_label = min(labelcounts)[1]
        al_idx = self.label_dict[min_label]
        learner = self.learners[al_idx]
        return learner



from typing import Optional

import numpy as np # type: ignore

def get_random_generator(
        rng: Optional[np.random.Generator] = None) -> np.random.Generator: # type: ignore
    if rng is not None:
        return rng # type: ignore
    return np.random.default_rng() # type: ignore



from dataclasses import dataclass
import itertools
import typing as ty
from abc import ABC, abstractmethod
from collections import OrderedDict
from os import PathLike
from typing import (
    Any,
    Dict,
    FrozenSet,
    Generic,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from ..estimation.base import AbstractEstimator, Estimate
from ..typehints import LT
from .experiments import ClassificationExperiment
from .plotter import ExperimentPlotter
from .statistics import DatasetStats
from instancelib.analysis.base import MulticlassModelMetrics

from dataclasses import dataclass
from .statistics import LabelALStatistics, ALStats

_T = TypeVar("_T")
_V = TypeVar("_V")
_W = TypeVar("_W")


def smooth_similar(
    xs: Union[npt.NDArray[Any], Sequence[_T]], ys: Union[npt.NDArray[Any], Sequence[_V]]
) -> Tuple[Sequence[_T], Sequence[_V]]:
    assert len(xs) == len(ys), f"len(xs={len(xs)}) != len(ys={len(ys)})"
    x_smoothed: List[_T] = list()
    y_smoothed: List[_V] = list()
    previous_y: Optional[_V] = None
    for x, y in zip(xs, ys):
        if previous_y != y:
            x_smoothed.append(x)
            y_smoothed.append(y)
            previous_y = y
    return x_smoothed, y_smoothed


def smooth_similar3(
    xs: Union[npt.NDArray[Any], Sequence[_T]],
    ys: Union[npt.NDArray[Any], Sequence[_V]],
    zs: Union[npt.NDArray[Any], Sequence[_W]],
) -> Tuple[Sequence[_T], Sequence[_V], Sequence[_W]]:
    assert (
        len(xs) == len(ys) == len(zs)
    ), f"len(xs={len(xs)}) != len(ys={len(ys)}) != len(zs={len(zs)})"
    x_smoothed: List[_T] = list()
    y_smoothed: List[_V] = list()
    z_smoothed: List[_W] = list()
    previous_y: Optional[_V] = None
    previous_z: Optional[_W] = None
    for x, y, z in zip(xs, ys, zs):
        if previous_y != y or previous_z != z:
            x_smoothed.append(x)
            y_smoothed.append(y)
            z_smoothed.append(z)
            previous_y = y
            previous_z = z
    return x_smoothed, y_smoothed, z_smoothed


class ClassificationPlotter(ExperimentPlotter[LT], Generic[LT]):
    dataset_name: str
    main_stats: ty.OrderedDict[int, Mapping[str, ALStats]]
    label_stats: ty.OrderedDict[int, Mapping[str, Mapping[LT, LabelALStatistics]]]
    metric_stats: ty.OrderedDict[int, Mapping[str, MulticlassModelMetrics]]
    stop_results: ty.OrderedDict[int, Mapping[str, Mapping[str, bool]]]

    def __init__(self, dataset_name: str = "") -> None:
        self.dataset_name = dataset_name

        self.dataset_stats = OrderedDict()
        self.metric_stats = OrderedDict()
        self.main_stats = OrderedDict()
        self.stop_results = OrderedDict()
        self.label_stats = OrderedDict()
        self.it = 0
        self.it_axis: List[int] = list()

    def update(
        self,
        exp_iterator: ClassificationExperiment[Any, Any, Any, Any, Any, LT],
        stop_result: Mapping[str, Mapping[str, bool]],
    ) -> None:
        self.it = exp_iterator.it
        self.it_axis.append(self.it)
        self.metric_stats[self.it] = exp_iterator.metrics
        self.label_stats[self.it] = exp_iterator.label_stats
        self.main_stats[self.it] = exp_iterator.main_stats
        self.stop_results[self.it] = exp_iterator.stop_result

    def print_last_stats(self) -> None:
        return

    def metric_names(self, model_name: str) -> FrozenSet[str]:
        return frozenset(["f1", "recall", "precision"])

    @property
    def model_names(self) -> FrozenSet[str]:
        if self.metric_stats:
            return frozenset(self.main_stats[self.it].keys())
        return frozenset()

    @property
    def criterion_names(self) -> FrozenSet[Tuple[str, str]]:
        if self.stop_results:
            return frozenset(
                [
                    (l_name, l_stop)
                    for l_name, l_stops in self.stop_results[self.it].items()
                    for l_stop in l_stops.keys()
                ]
            )
        return frozenset()

    def _plot_metrics(
        self,
        included_models: Optional[Sequence[str]] = None,
        included_metrics: Optional[Sequence[str]] = None,
        included_labels: Optional[Sequence[LT]] = None,
    ) -> None:
        model_selection = (
            included_models if included_models is not None else list(self.model_names)
        )
        i = 0
        for model in model_selection:
            metric_selection = (
                included_metrics
                if included_metrics is not None
                else list(self.metric_names(model))
            )
            label_selection = (
                included_labels
                if included_labels is not None
                else list(self.label_stats[self.it][model].keys())
            )
            effort_axis = self._effort_axis(model)

            combis = itertools.product(metric_selection, label_selection)
            for (metric, label) in combis:
                metrics = {
                    "f1": np.array(
                        [self.metric_stats[it][model][label].f1 for it in self.it_axis]
                    )
                    * 100,
                    "recall": np.array(
                        [
                            self.metric_stats[it][model][label].recall
                            for it in self.it_axis
                        ]
                    )
                    * 100,
                    "precision": np.array(
                        [
                            self.metric_stats[it][model][label].precision
                            for it in self.it_axis
                        ]
                    )
                    * 100,
                    "accuracy": np.array(
                        [
                            self.metric_stats[it][model][label].accuracy
                            for it in self.it_axis
                        ]
                    )
                    * 100,
                }
                color = f"C{i}"
                plt.plot(
                    effort_axis,
                    metrics[metric],
                    linestyle="-.",
                    label=f"{model} -- {metric} -- {label}",
                    color=color,
                )
                i += 1

    def _effort_axis(self, model: str) -> npt.NDArray[Any]:
        effort_axis = np.array(
            [
                self.main_stats[it][model].labeled / self.main_stats[it][model].dataset
                for it in self.it_axis
            ]
        )
        return effort_axis * 100

    def _graph_setup(self, simulation=True) -> None:
        model_name = next(iter(self.model_names))
        dataset_size = self.main_stats[self.it][model_name].dataset

        plt.xlabel(f"% annotated")
        plt.ylabel("result of the metric")

        if self.dataset_name:
            plt.title(
                f"Run on the `{self.dataset_name}` dataset with size {dataset_size}."
            )
        else:
            plt.title(f"Run on a dataset with size {dataset_size}.")

    def show(
        self,
        x_lim: Optional[float] = None,
        y_lim: Optional[float] = None,
        included_models: Optional[Sequence[str]] = None,
        included_metrics: Optional[Sequence[str]] = None,
        included_labels: Optional[Sequence[LT]] = None,
        filename: "Optional[PathLike[str]]" = None,
    ):
        self._graph_setup(simulation=False)
        self._plot_metrics(included_models, included_metrics, included_labels)
        # self._plot_stop_criteria()
        # self._set_axes(x_lim, y_lim)
        plt.legend(bbox_to_anchor=(1.05, 1), loc="upper left")
        if filename is not None:
            plt.savefig(filename, bbox_inches="tight")



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generic, TypeVar, Optional, Callable

_T = TypeVar("_T")


class ObjectFactory:
    def __init__(self):
        self.builders = {}

    def get_constructor(self, key: Any) -> Any:
        builder = self.builders[key]
        if isinstance(builder, ObjectBuilder):
            return builder.constructor            
        raise KeyError(
            f"The requested key '{key}' has no retrievable constructor")

    def register_constructor(self, key: Any, constructor: Any) -> None:
        builder = ObjectBuilder[Any](constructor)
        self.register_builder(key, builder)

    def register_builder(self, key: Any, builder: AbstractBuilder):
        builder.register_factory(self)
        self.builders[key] = builder

    def create(self, key: Any, **kwargs):
        builder = self.builders.get(key)
        if not builder:
            raise NotImplementedError(
                "The module '{}' is not registered".format(key))
        return builder(**kwargs)

    def attach(self, factory: ObjectFactory):
        for key, builder in factory.builders.items():
            self.register_builder(key, builder)


class AbstractBuilder(ABC):
    _factory: ObjectFactory
    def __init__(self):
        self._factory = ObjectFactory()

    def register_factory(self, factory: ObjectFactory):
        self._factory = factory

    @abstractmethod
    def __call__(self, **kwargs): # type: ignore
        raise NotImplementedError


class ObjectBuilder(AbstractBuilder, Generic[_T]):
    def __init__(self, constructor: Callable[..., _T]) -> None:
        super().__init__()
        self.constructor = constructor

    def __call__(self, **kwargs) -> _T:
        return self.constructor(**kwargs)



import collections
from typing import Any, Deque, Generic

from instancelib.utils.func import all_equal

from ..activelearning import ActiveLearner
from ..analysis.analysis import process_performance
from ..typehints import LT
from .base import AbstractStopCriterion

class AllDocsCriterion(AbstractStopCriterion[LT], Generic[LT]):
    def __init__(self) -> None:
        self.remaining = 2000

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        self.remaining = len(learner.env.unlabeled)
    
    @property
    def stop_criterion(self) -> bool:
        return self.remaining <= 0

class DocCountStopCritertion(AbstractStopCriterion[LT], Generic[LT]):
    def __init__(self, max_docs: int):
        self.max_docs = max_docs
        self.doc_count = 0
    
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        self.doc_count = len(learner.env.labeled)
    
    @property
    def stop_criterion(self) -> bool:
        return self.doc_count >= self.max_docs

class SameStateCount(AbstractStopCriterion[LT], Generic[LT]):
    def __init__(self, label: LT, same_state_count: int):
        self.label = label
        self.same_state_count = same_state_count
        self.pos_history: Deque[int] = collections.deque()
        self.has_been_different = False

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]):
        performance = process_performance(learner, self.label)
        self.add_count(len(performance.true_positives))

    def add_count(self, value: int) -> None:
        if len(self.pos_history) > self.same_state_count:
            self.pos_history.pop()
        if self.pos_history and not self.has_been_different:
            previous_value = self.pos_history[0]
            if previous_value != value:
                self.has_been_different = True
        self.pos_history.appendleft(value)

    @property
    def count(self) -> int:
        return self.pos_history[0]

    @property
    def same_count(self) -> bool:
        return all_equal(self.pos_history)

    @property
    def stop_criterion(self) -> bool:
        if len(self.pos_history) < self.same_state_count:
            return False 
        return self.has_been_different and self.same_count


class AprioriRecallTarget(AbstractStopCriterion[LT]):
    def __init__(self, pos_label: LT,  target: float = 0.95):
        self.target = target
        self.stopped = False
        self.pos_label = pos_label

    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        if not self.stopped:
            n_pos_truth = learner.env.truth.document_count(self.pos_label)
            n_pos_now = learner.env.labels.document_count(self.pos_label)
            self.stopped = n_pos_now / n_pos_truth >= self.target

    @property
    def stop_criterion(self) -> bool:
        return self.stopped



class FractionStopCritertion(AbstractStopCriterion[LT], Generic[LT]):
    read_fraction: float
    target_fraction: float

    def __init__(self, target_fraction: float):
        self.target_fraction = target_fraction
        self.read_fraction = 0.0
    
    def update(self, learner: ActiveLearner[Any, Any, Any, Any, Any, LT]) -> None:
        self.read_fraction = len(learner.env.labeled) / len(learner.env.dataset)
    
    @property
    def stop_criterion(self) -> bool:
        return self.read_fraction >= self.target_fraction